"""Local neuroglancer instance with a shell-like interface."""
# stdlib
import argparse
import atexit
import contextlib
import datetime
import logging
import os.path
import stat
import sys
import textwrap
from functools import partial, wraps
from typing import Generator

# externals
import numpy as np

# internals
from ngtools.local.console import ActionHelpFormatter, Console
from ngtools.local.fileserver import LocalFileServer, StaticFileHandler
from ngtools.local.handlers import LincHandler, LutHandler
from ngtools.local.termcolors import bformat
from ngtools.optionals import try_from_import
from ngtools.scene import Scene
from ngtools.shaders import pretty_colormap_list
from ngtools.utils import NG_URLS, find_available_port

# optionals
try:
    import neuroglancer as ng
except ImportError:
    import ngtools._nglite as ng

ng_bind_address, ng_stop_server = try_from_import(
    "neuroglancer.server", ["set_server_bind_address", "stop"],
    fallback="ngtools._nglite.server",
)

# unix-specific imports
try:
    import grp
    import pwd
except ImportError:
    pwd = grp = None


LOG = logging.getLogger(__name__)


def action(needstate: bool = False) -> callable:
    """
    Decorate a neuroglancer action.

    These actions can thereby be triggered by argparse.
    """
    if callable(needstate):
        return action()(needstate)

    def decorator(func: callable) -> callable:

        @wraps(func)
        def wrapper(
            self: "LocalNeuroglancer", *args: tuple, **kwargs: dict
        ) -> callable:
            args = list(args)
            if args and isinstance(args[0], argparse.Namespace):
                if len(args) > 1:
                    raise ValueError(
                        'Only one positional argument accepted when an '
                        'action is applied to an argparse object')
                parsedargs = vars(args.pop(0))
                parsedargs.update(kwargs)
                parsedargs.pop('func', None)
                kwargs = parsedargs

            if needstate and not kwargs.get('state', None):
                with self.viewer.txn(overwrite=True) as state:
                    kwargs['state'] = state
                    result = func(self, *args, **kwargs)
                return result
            else:
                result = func(self, *args, **kwargs)
                return result

        return wrapper

    return decorator


def state_action(name: str) -> callable:
    """Generate a state action that wraps a scene action."""

    @action(needstate=True)
    @wraps(getattr(Scene, name))
    def func(
        self: "LocalNeuroglancer", *args, state: ng.ViewerState, **kwargs
    ) -> object | None:
        # build scene
        scene = Scene(state.to_json(), stdio=self.console.stdio)
        # if loader, pass fileserver
        if name in ("load", "shader"):
            kwargs["fileserver"] = self.get_fileserver_url()
        # run action
        scene_fn = getattr(scene, name)
        out = scene_fn(*args, **kwargs)
        # save state
        for key in scene.to_json().keys():
            val = getattr(scene, key)
            setattr(state, key, val)
        return out

    return func


def ensure_list(x: object) -> list:
    """Ensure that an object is a list. Make one if needed."""
    if isinstance(x, np.ndarray):
        x = x.tolist()
    if not isinstance(x, (list, tuple)):
        x = [x]
    return list(x)


class OSMixin:
    """Operating system actions."""

    @action
    def cd(self, path: str) -> str:
        """Change directory."""
        os.chdir(os.path.expanduser(path))
        return os.getcwd()

    @action
    def ls(self, path: str, **kwargs) -> list[str]:
        """List files."""
        long = kwargs.pop("long", False)
        hidden = kwargs.pop("hidden", False)

        if long:
            files = os.scandir(os.path.expanduser(path))
            files = sorted(files, key=lambda x: x.name)
            keys = ("perms", "nlinks", "owners", "groups", "sizes",
                    "months", "days", "times", "names")
            mystat = dict([(key, []) for key in keys])
            for entry in files:
                if not hidden and entry.name[:1] == ".":
                    continue
                lstat = entry.stat(follow_symlinks=False)
                mystat["perms"].append(stat.filemode(lstat.st_mode))
                mystat["nlinks"].append(lstat.st_nlink)
                if pwd:
                    mystat["owners"].append(pwd.getpwuid(lstat.st_uid).pw_name)
                else:
                    mystat["owners"].append(lstat.st_uid)
                if grp:
                    mystat["groups"].append(grp.getgrgid(lstat.st_gid)[0])
                else:
                    mystat["groups"].append("")
                mystat["sizes"].append(lstat.st_size)
                date = datetime.datetime.fromtimestamp(lstat.st_mtime)
                mystat["months"].append(date.strftime("%b"))
                mystat["days"].append(date.strftime("%d"))
                if date.year != datetime.datetime.today().year:
                    mystat["times"].append(date.strftime("%Y"))
                else:
                    mystat["times"].append(date.strftime("%H:%M"))
                if entry.is_symlink():
                    fname = bformat.bold(bformat.fg.magenta(entry.name))
                    fname += " -> " + os.readlink(entry.path)
                elif entry.is_dir():
                    fname = bformat.bold(bformat.fg.blue(entry.name))
                elif "x" in mystat["perms"][-1]:
                    fname = bformat.bold(bformat.fg.green(entry.name))
                else:
                    fname = entry.name
                mystat["names"].append(fname)

            colsizes = [max(len(str(y)) for y in x) for x in mystat.values()]
            for values in zip(*mystat.values()):
                for key, value, colsize in zip(keys, values, colsizes):
                    align = (
                        ">" if key in ("nlinks", "sizes", "days", "times")
                        else ""
                    )
                    fmt = "{:" + align + str(colsize) + "s}"
                    print(fmt.format(str(value)), end=" ")
                print("")
        else:
            files = os.listdir(os.path.expanduser(path))
            if not hidden:
                files = [file for file in files if file[:1] != "."]
            print(*files)
        return files

    @action
    def pwd(self) -> str:
        """Path to working directory."""
        self.console.print(os.getcwd())
        return os.getcwd()

    @action
    def stdin(self, stdin: str | None = None) -> str:
        """Input stream."""
        if stdin:
            if stdin == "sys":
                stdin = sys.stdin
            elif stdin[:4] == "sys.":
                parts = stdin.split(".")[1:]
                stdin = sys
                for part in parts:
                    stdin = getattr(stdin, part)
            self.console.stdio.stdin = stdin
        return self.console.stdio.stdin

    @action
    def stdout(self, stdout: str | None = None) -> str:
        """Output stream."""
        if stdout:
            if stdout == "sys":
                stdout = sys.stdout
            elif stdout[:4] == "sys.":
                parts = stdout.split(".")[1:]
                stdout = sys
                for part in parts:
                    stdout = getattr(stdout, part)
            self.console.stdio.stdout = stdout
        return self.console.stdio.stdout

    @action
    def stderr(self, stderr: str | None = None) -> str:
        """Error stream."""
        if stderr:
            if stderr == "sys":
                stderr = sys.stderr
            elif stderr[:4] == "sys.":
                parts = stderr.split(".")[1:]
                stderr = sys
                for part in parts:
                    stderr = getattr(stderr, part)
            self.console.stdio.stderr = stderr
        return self.console.stdio.stderr


class LocalNeuroglancer(OSMixin):
    """
    A local instance of neuroglancer that can launch its own local fileserver.

    It also comes with a shell-like interface that allows loading,
    unloading, applying transforms, etc.
    """

    def __init__(
        self,
        port: int = 0,
        ip: str = "localhost",
        token: int = 1,
        fileserver: bool | int | LocalFileServer = True,
        **console_kwargs,
    ) -> None:
        """
        Parameters
        ----------
        port : int
            Port to use.
        ip : str
            IP to use.
        token : str
            Unique id for the instance.
        fileserver : bool | int | LocalFileServer
            Whether to run a local fileserver.
            If an int, should be the port to use.
        debug : bool
            Print full trace when an error is encountered.
        """
        # Setup neuroglancer instance
        port, ip = find_available_port(port, ip)
        ng_bind_address(str(ip), str(port))
        self.viewer = ng.Viewer(token=str(token))
        # self.viewer.shared_state.add_changed_callback(self.on_state_change)

        ip = self.get_viewer_url().split("://")[1].split(":")[0]

        # Add specific handlers
        if fileserver is not False:
            if not isinstance(fileserver, LocalFileServer):
                if fileserver is True:
                    fileserver_port = 0
                else:
                    fileserver_port = int(fileserver)
                fileserver = LocalFileServer(fileserver_port, ip)

            fileserver.app.handlers.extend([
                (r"^/local/(.*)", StaticFileHandler),
                (r"^/lut/([^/]+)/(.*)", LutHandler),
                (r"^/linc/(.*)", LincHandler),
            ])
        self.fileserver = fileserver

        # Setup console
        self.console = self._make_console(**console_kwargs)
        atexit.register(self._cleanup)

    def _cleanup(self) -> None:
        ng_stop_server()  # do I need this?
        atexit.unregister(self._cleanup)

    def __del__(self) -> None:
        self._cleanup()

    def txn(self) -> contextlib.AbstractContextManager[ng.ViewerState]:
        """
        Context manager that returns the underlying neuroglancer state.
        Transactions are written into the viewer at exit.
        """
        return self.viewer.txn()

    @contextlib.contextmanager
    def _scene(self) -> Generator[Scene, None, None]:
        with self.viewer.txn() as state:
            # Build and yield scene
            scene = Scene(state.to_json(), stdio=self.console.stdio)
            yield scene
            # Save scene's state into viewer's state
            for key in scene.to_json().keys():
                val = getattr(scene, key)
                setattr(state, key, val)

    def scene(self) -> contextlib.AbstractContextManager[Scene]:
        """
        Context manager that returns the underlying state as a `Scene`.
        It writes all changes to the scene back into the state at exit.
        """
        return self._scene()

    def get_server_url(self) -> str:
        """URL of the neuroglancer server."""
        return ng.server.get_server_url().rstrip("/") + "/"

    def get_viewer_url(self) -> str:
        """URL of the viewer."""
        return self.viewer.get_viewer_url().rstrip("/") + "/"

    def get_fileserver_url(self) -> str:
        """URL of the neuroglancer server."""
        return self.fileserver.get_url().rstrip("/") + "/"

    # ==================================================================
    #
    #                   COMMANDLINE APPLICATION
    #
    # ==================================================================

    def await_input(self) -> None:
        """Launch shell-like interface."""
        return self.console.await_input()
        # self._await_input = Thread(target=self.console.await_input)
        # self._await_input.start()

    def _make_console(self, **kwargs) -> Console:
        kwargs.setdefault('max_choices', 4)
        mainparser = Console('', **kwargs)
        parsers = mainparser.add_subparsers()
        F = dict(formatter_class=ActionHelpFormatter)

        def add_parser(cmd, *args, **kwargs):  # noqa: ANN001, ANN202
            # We define long descriptions in the _clihelp class at the
            # end of this file.
            description = kwargs.get("description", None)
            if description is None:
                description = getattr(_clihelp, cmd, None)
            if description is None:
                description = kwargs.get("help", None)
            if description:
                description = description.rstrip()
                description += "\n\n"
                description += _clihelp.header_attributes
            kwargs.setdefault('description', description)
            return parsers.add_parser(cmd, *args, **kwargs, **F)

        # --------------------------------------------------------------
        #   HELP
        # --------------------------------------------------------------
        _ = add_parser('help', help='Display help')
        _.set_defaults(func=self.help)
        _.add_argument(
            dest='action', nargs='?', help='Command for which to display help')

        # --------------------------------------------------------------
        #   LOAD
        # --------------------------------------------------------------
        _ = add_parser('load', help='Load a file')
        _.set_defaults(func=self.load)
        _.add_argument(
            dest='filename', nargs='+', metavar='FILENAME',
            help='Filename(s) with protocols')
        _.add_argument(
            '--name', '-n', nargs='+', help='A name for the image layer')
        _.add_argument(
            '--transform', '-t', nargs='+', help='Apply a transform')
        _.add_argument(
            '--shader', '-s', help='Apply a shader')

        # --------------------------------------------------------------
        #   UNLOAD
        # --------------------------------------------------------------
        _ = add_parser('unload', help='Unload a file')
        _.set_defaults(func=self.unload)
        _.add_argument(
            dest='layer', nargs='+', metavar='LAYER',
            help='Name(s) of layer(s) to unload')

        # --------------------------------------------------------------
        #   RENAME
        # --------------------------------------------------------------
        _ = add_parser('rename', help='Rename a file')
        _.set_defaults(func=self.rename)
        _.add_argument(
            dest='src', metavar='SOURCE', help='Current layer name')
        _.add_argument(
            dest='dst', metavar='DEST', help='New layer name')

        # --------------------------------------------------------------
        #   WORLD AXES
        # --------------------------------------------------------------
        _ = add_parser('world_axes', help='Rename native axes')
        _.set_defaults(func=self.world_axes)
        _.add_argument(
            dest='axes', metavar='DEST', nargs="*", help='New axis names')
        _.add_argument(
            '--destination', '--dst', '-d', metavar='DEST', nargs="+",
            help='New axis names')
        _.add_argument(
            '--source', '--src', '-s', metavar='SOURCE', nargs="*",
            help='Native axis names')
        _.add_argument(
            '--print', '-p', action="store_true", help='Print result')

        # --------------------------------------------------------------
        #   RENAME AXES
        # --------------------------------------------------------------
        _ = add_parser('rename_axes', help='Rename axes')
        _.set_defaults(func=self.rename_axes)
        _.add_argument(
            dest='dst', metavar='DEST', nargs="+", help='New axis names')
        _.add_argument(
            '--destination', '--dst', '-d', metavar='DEST', nargs="+",
            dest="dst", help='New axis names')
        _.add_argument(
            '--source', '--src', '-s', metavar='SOURCE', nargs="*",
            dest="src", help='Old axis names')
        _.add_argument(
            '--layer', '-l', nargs="*", help='Layer(s) to rename')

        # --------------------------------------------------------------
        #   SPACE
        # --------------------------------------------------------------
        MODES = ("radio", "neuro", "default")
        _ = add_parser('space', help='Cross-section orientation')
        _.set_defaults(func=self.space)
        _.add_argument(
            dest='mode', nargs="?", help='New axis names', choices=MODES)
        _.add_argument(
            '--layer', '-l',
            help='If the name of a layer, align the cross-section with its '
                 'voxel grid. If "world", align the cross section with the '
                 'canonical axes.')

        # --------------------------------------------------------------
        #   TRANSFORM
        # --------------------------------------------------------------
        _ = add_parser('transform', help='Apply a transform')
        _.set_defaults(func=self.transform)
        _.add_argument(
            dest='transform', nargs='*', metavar='TRANSFORM',
            help='Path to transform file or flattened transformation '
                 'matrix (row major)')
        _.add_argument(
            '--layer', '-l', nargs='+',
            help='Name(s) of layer(s) to transform')
        _.add_argument(
            '--inv', '-i', action='store_true', default=False,
            help='Invert the transform before applying it')
        _.add_argument(
            '--reset', '-r', action='store_true', default=False,
            help="Reset the layer's transform before applying the new one")
        _.add_argument(
            '--mov', '-m', help='Moving image (required by some formats)')
        _.add_argument(
            '--fix', '-f', help='Fixed image (required by some formats)')

        # Save current transform
        _ = add_parser('save_transform', help='Save the current transform')
        _.set_defaults(func=self.save_transform)
        _.add_argument(
            dest='output', nargs='*', help='Path to output transform.')
        _.add_argument(
            '--layer', '-l', nargs='+', help='Name(s) of layer(s) to save')
        _.add_argument(
            '--format', '-f', help='Format to save into')

        # --------------------------------------------------------------
        #   AXIS MODE
        # --------------------------------------------------------------
        _ = add_parser(
            'channel_mode', help='Change the way a dimension is interpreted')
        _.set_defaults(func=self.channel_mode)
        _.add_argument(
            dest='mode',
            choices=('local', 'channel', 'global'),
            help='How to interpret the channel (or another) axis')
        _.add_argument(
            '--layer', '-l', nargs='+', default=None,
            help='Name(s) of layer(s) to transform')
        _.add_argument(
            '--dimension', '-d', nargs='+', default=['c'],
            help='Name(s) of axes to transform')

        # --------------------------------------------------------------
        #   SHADER
        # --------------------------------------------------------------
        _ = add_parser('shader', help='Apply a shader')
        _.set_defaults(func=self.shader)
        _.add_argument(
            dest='shader', metavar='SHADER',
            help='Shader name or GLSL shader code')
        _.add_argument(
            '--layer', '-l', nargs='+',
            help='Layer(s) to apply shader to')
        _.add_argument(
            '--layer-type', '-t', nargs='+',
            help='Layer type(s) to apply shader to')

        # --------------------------------------------------------------
        #   DISPLAY
        # --------------------------------------------------------------
        _ = add_parser('display', help='Dimensions to display')
        _.set_defaults(func=self.display)
        _.add_argument(
            dest='dimensions', nargs='*', metavar='DIMENSIONS',
            help='Dimensions to display')

        # --------------------------------------------------------------
        #   LAYOUT
        # --------------------------------------------------------------
        LAYOUTS = ["xy", "yz", "xz", "xy-3d", "yz-3d", "xz-3d", "4panel", "3d"]
        _ = add_parser('layout', help='Layout')
        _.set_defaults(func=self.layout)
        _.add_argument(
            dest='layout', nargs='*', choices=LAYOUTS, metavar='LAYOUT',
            help='Layout')
        _.add_argument(
            '--stack', '-s', choices=("row", "column"), help="Stack direction")
        _.add_argument(
            '--layer', '-l', nargs='*', help="Layer(s) to include")
        _.add_argument(
            '--flex', type=float, default=1, help="Flex")
        _.add_argument(
            '--append', '-a', type=int, nargs='*',
            help="Append to existing (nested) layout")
        _.add_argument(
            '--assign', '-x', type=int, nargs='*',
            help="Assign into existing (nested) layout")
        _.add_argument(
            '--insert', '-i', type=int, nargs='+',
            help="Insert in existing (nested) layout")
        _.add_argument(
            '--remove', '-r', type=int, nargs='+',
            help="Remove from an existing (nested) layout")
        _.add_argument(
            '--row', dest="stack", action="store_const", const="row",
            help="Alias for `--stack row`")
        _.add_argument(
            '--column', "--col", dest="stack",
            action="store_const", const="column",
            help="Alias for `--stack column`")

        # --------------------------------------------------------------
        #   STATE
        # --------------------------------------------------------------
        _ = add_parser('state', help='Return the viewer\'s state')
        _.set_defaults(func=self.state)
        _.add_argument(
            '--no-print', action='store_false', default=True, dest='print',
            help='Do not print the state.')
        _.add_argument(
            '--save', '-s', help='Save JSON state to this file.')
        _.add_argument(
            '--load', '-l',
            help='Load JSON state from this file. '
                 'Can also be a JSON string or a URL.')
        _.add_argument(
            '--url', '-u', action='store_true', default=False,
            help='Load (or print) the url form of the state')
        _.add_argument(
            '--open', '-o', action='store_true', default=False,
            help='Open the url (if `--url`) or viewer (otherwise)')
        _.add_argument(
            '--instance', '-i', choices=list(NG_URLS.keys()),
            help='Link to this neuroglancer instance')

        # --------------------------------------------------------------
        #   POSITION
        # --------------------------------------------------------------
        _ = add_parser('move', help='Move cursor', aliases=("position",))
        _.set_defaults(func=self.move)
        _.add_argument(
            dest='coord', nargs='*', metavar='COORD', type=float,
            help='Cursor coordinates. If None, print current one.')
        _.add_argument(
            '--dimensions', '-d', nargs='+', default=None,
            help='Axis name for each coordinate (can be compact)')
        _.add_argument(
            '--unit', '-u',
            help='Coordinates are expressed in this unit')
        _.add_argument(
            '--absolute', '--abs', '-a', action='store_true', default=False,
            help='Move to absolute position, rather than relative to current')
        _.add_argument(
            '--reset', '-r', action='store_true', default=False,
            help='Reset coordinates to zero')

        # --------------------------------------------------------------
        #   ZOOM
        # --------------------------------------------------------------
        _ = add_parser('zoom', help='Zoom by a factor [default: x2]')
        _.set_defaults(func=self.zoom)
        _.add_argument(
            dest='factor', nargs='?', type=float, default=2.0,
            help='Zoom factor.')
        _.add_argument(
            '--reset', '-r', action='store_true', default=False,
            help='Reset zoom level to default.')

        _ = add_parser('unzoom', help='Unzoom by a factor [default: ÷2]')
        _.set_defaults(func=self.unzoom)
        _.add_argument(
            dest='factor', nargs='?', type=float, default=2.0,
            help='Inverse zoom factor.')
        _.add_argument(
            '--reset', '-r', action='store_true', default=False,
            help='Reset zoom level to default.')

        # --------------------------------------------------------------
        #   ZORDER
        # --------------------------------------------------------------
        _ = add_parser('zorder', help='Reorder layers')
        _.set_defaults(func=self.zorder)
        _.add_argument(
            dest='layer', nargs='+', metavar='LAYER',
            help='Layer(s) name(s)')
        _.add_argument(
            '--up', '-u', '-^', action='count', default=0,
            help='Move upwards')
        _.add_argument(
            '--down', '-d', '-v', action='count', default=0,
            help='Move downwards')

        # --------------------------------------------------------------
        #   NAVIGATION
        # --------------------------------------------------------------
        _ = add_parser('cd', help='Change directory')
        _.set_defaults(func=self.cd)
        _.add_argument(dest='path', metavar='PATH')

        _ = add_parser('ls', help='List files')
        _.set_defaults(func=self.ls)
        _.add_argument(dest='path', nargs='?', default='.', metavar='PATH')
        _.add_argument('--long', '-l', action="store_true")
        _.add_argument('--hidden', '-a', action="store_true")

        _ = add_parser('ll', help='List files (long form)')
        _.set_defaults(func=partial(self.ls, long=True, hidden=True))
        _.add_argument(dest='path', nargs='?', default='.', metavar='PATH')

        _ = add_parser('pwd', help='Path to working directory')
        _.set_defaults(func=self.pwd)

        _ = add_parser('stdin', help='Set input stream')
        _.set_defaults(func=self.stdin)
        _.add_argument(dest='stdin', metavar='FILE')
        _ = add_parser('stdout', help='Set output stream')
        _.set_defaults(func=self.stdout)
        _.add_argument(dest='stdout', metavar='FILE')
        _ = add_parser('stderr', help='Set error stream')
        _.set_defaults(func=self.stderr)
        _.add_argument(dest='stderr', metavar='FILE')

        # --------------------------------------------------------------
        #   EXIT
        # --------------------------------------------------------------
        _ = add_parser('exit', aliases=['quit'], help='Exit neuroglancer')
        _.set_defaults(func=self.exit)
        return mainparser

    # ==================================================================
    #
    #                              ACTIONS
    #
    # ==================================================================

    load = state_action("load")
    unload = state_action("unload")
    rename = state_action("rename")
    world_axes = state_action("world_axes")
    rename_axes = state_action("rename_axes")
    space = state_action("space")
    display = state_action("display")
    transform = state_action("transform")
    save_transform = state_action("save_transform")
    channel_mode = state_action("channel_mode")
    move = state_action("move")
    zoom = state_action("zoom")
    unzoom = state_action("unzoom")
    shader = state_action("shader")
    layout = state_action("change_layout")
    zorder = state_action("zorder")
    state = state_action("state")

    @action
    def help(self, action: str | None = None) -> None:
        """
        Display help.

        Parameters
        ----------
        action : str
            Action for which to display help
        """
        if action:
            self.console.parse_args([action, '--help'])
        else:
            self.console.parse_args(['--help'])

    @action
    def exit(self) -> None:
        """Exit gracefully."""
        raise SystemExit

    # ==================================================================
    #
    #                            CALLBACKS
    #
    # ==================================================================

    # def on_state_change(self) -> None:
    #     old_state = self.saved_state
    #     with self.viewer.txn() as state:
    #         self.saved_state = state
    #         for layer in state.layers:
    #             name = layer.name
    #             layer = layer.layer
    #             if isinstance(layer.source, ng.SkeletonSource):
    #                 if (layer.shaderControls.nbtracts
    #                         != old_state.shaderControls.nbtracts):
    #                     self.update_tracts_nb(name)

    # def update_tracts_nb(self, name: str) -> None:
    #     with self.viewer.txn() as state:
    #         layer = state.layers[name]
    #         print(type(layer))
    #         layer.source.max_tracts = layer.shaderControls.nbtracts
    #         layer.source._filter()
    #         layer.invalidate()


_clihelp = type('_clihelp', (object,), {})

b = bformat.bold
u = bformat.underline
i = bformat.italic
R = bformat.fg.red
G = bformat.fg.green
B = bformat.fg.blue

_clihelp.header_attributes = textwrap.dedent(
    f"{b('Arguments')}\n{b('----------')}\n"
)

_clihelp.load = textwrap.dedent(
f"""
Load a file, which can be local or remote.

{b("Paths and URLs")}
{b("--------------")}
Each path or url may be prepended by:

1)  A layer type protocol that indicates the kind of object that the file
    contains.
    Examples: {b("volume://")}, {b("labels://")}, {b("tracts://")}.

2)  A format protocol that indicates the exact file format.
    Examples: {b("nifti://")}, {b("zarr://")}, {b("mgh://")}.

3)  An access protocol that indicates the protocol used to access the files.
    Examples: {b("https://")}, {b("s3://")}, {b("dandi://")}.

All of these protocols are optional. If absent, a guess is made using the
file extension.

{b("Examples")}
{b("--------")}

. Absolute path to local file:  {b("/absolute/path/to/mri.nii.gz")}
. Relative path to local file:  {b("relative/path/to/mri.nii.gz")}
. Local file with format hint:  {b("mgh://relative/path/to/linkwithoutextension")}
. Remote file:                  {b("https://url.to/mri.nii.gz")}
. Remote file with format hint: {b("zarr://https://url.to/filewithoutextension")}
. File on dandiarchive:         {b("dandi://dandi/<dandiset>/sub-<id>/path/to/file.ome.zarr")}

{b("Layer names")}
{b("-----------")}
Neuroglancer layers are named. The name of the layer can be specified with
the {b("--name")} option. Otherwise, the base name of the file is used (that
is, without the folder hierarchy).

If multiple files are loaded {i("and")} the {b("--name")} option is used, then there
should be as many names as files.

{b("Transform")}
{b("---------")}
A spatial transform (common to all files) can be applied to the loaded
volume. The transform is specified with the {b("--transform")} option, which
can be a flattened affine matrix (row major) or the path to a transform file.
Type {b("help transform")} for more information.

{b("Shader")}
{b("------")}
A shader (= colormap, common to all files) can be applied to the loaded
volume. The shader is specified with the {b("--shader")} option, which
can be the name of a colormap, the path to a LUT file, or a snippet of
GLSL code. Type {b("help shader")} for more information.
"""  # noqa: E122, E501
)

_clihelp.unload = "Unload layers"

_clihelp.rename = "Rename a layer"

_clihelp.shader = textwrap.dedent(
f"""
Applies a colormap, or a more advanced shading function to all or some of
the layers.

The input can also be the path to a (local or remote) freesurfer LUT file.

{b("List of builtin colormaps")}
{b("-------------------------")}
"""  # noqa: E122
) + textwrap.indent(pretty_colormap_list(), ' ')

_clihelp.zorder = textwrap.dedent(
f"""
Modifies the z-order of the layers.

In neuroglancer, layers are listed in the order they are loaded, with the
latest layer appearing on top of the other ones in the scene.

Counter-intuitively, the latest/topmost layer is listed at the bottom of
the layer list, while the earliest/bottommost layer is listed at the top of
the layer list. In this command, layers should be listed in their expected
z-order, top to bottom.

There are two ways of using this command:

1)  Provide the new order (top-to-bottom) of the layers

2)  Provide a positive ({b("--up")}) or negative ({b("--down")})
    number of steps by which to move the listed layers.
    In this case, more than one up or down step can be provided, using
    repeats of the option.
    Examples: {b("-vvvv")} moves downward 4 times
              {b("-^^^^")} moves upwards 4 times
"""  # noqa: E122, E501
)

_clihelp.display = textwrap.dedent(
f"""
Neuroglancer is quite flexible in the sense that it does not have a
predefined "hard" coordinate frame in which the data is shown. Instead,
the {R("red")}, {G("green")} and {B("blue")} "visual" axes can be
arbitrarily mapped to any of the existing "model" axes.

By default, most formats use ({b("x")}, {b("y")}, {b("z")}) as their
model axes, although they may not have predefined anatomical meaning.
NIfTI files do have an anatomical convention, under which axes
{R("x")}, {G("y")}, {B("z")} map to the {R("right")}, {G("anterior")}, {B("superior")} sides of the brain.

In order to show data in a frame that is standard in the neuroimaging
field, we therefore map the visual axes {R("red")}, {G("green")}, {B("blue")}
to the model axes {R("x")}, {G("y")}, {B("z")} when loading data in an
empty scene. This mapping is also enforced every time the command
{b("space")} is used.

The {b("display")} command can be used to assign a different set of
model axes to the visual axes.

{u("See also")}:  {b("space")}, {b("world_axes")}
"""  # noqa: E122, E501
)

_clihelp.world_axes = textwrap.dedent(
f"""
Neuroglancer is quite flexible in the sense that it does not have a
predefined "hard" coordinate frame in which the data lives. Instead,
arbitrary "model" axes can be defines, in terms of an affine transformation
of "native" axes.

By default, most formats use ({b("x")}, {b("y")}, {b("z")}) as their
model axes, although they may not have predefined anatomical meaning.
NIfTI files do have an anatomical convention, under which axes
{R("x")}, {G("y")}, {B("z")} map to the {R("right")}, {G("anterior")}, {B("superior")} sides of the brain.

This function allows native axes to replaces by more anatomically
meaningful names, by leveraging neuroglancer's transforms. In order
to allow these transforms to be undone, and new data to be loaded without
introducing conflicting coordinate frames, we store the mapping inside
local annotation layers named "__world_axes_native__" and
"__world_axes_current__".

{u("See also")}:  {b("space")}, {b("display")}

{b("Examples")}
{b("--------")}

. `world_axes right anterior superior` : x -> right, y -> anterior, z -> superior
. `world_axes ras`                     : x -> right, y -> anterior, z -> superior
. `world_axes u v w --src z y x`       : z -> u, y -> v, z -> w
"""  # noqa: E122, E501
)

_clihelp.space = textwrap.dedent(
f"""
This function rotates and orients the cross-section plane such that

. Visual axes are pointed according to some defined convention
  (radio or neuro)
. The cross section is aligned with either the model coordinate frame,
  or one of the layer's voxel space.

{u("Note:")} when this function is used, the displayed axes are always reset
to {R("x")}, {G("y")}, {B("z")} (or their world names if {b("world_axes")}) has been used.

{b("Examples")}
{b("--------")}

. `space radio`                 : Orients the cross section such that the {R("x")}
                                  axis points to the {b("left")} of the first quadrant
                                  (radiological convention).
. `space neuro`                 : Orients the cross section such that the {R("x")}
                                  axis points to the {b("right")} of the first quadrant
                                  (neurological convention).
. `space --layer <LAYER>`       : Aligns the cross-section with the voxels of
                                  the designated layer, while keeping the
                                  existing radio or neuro convention.
. `space --layer world`         : Aligns the cross-section with the canonical
                                  model space, while keeping the existing radio
                                  or neuro convention.
. `space radio --layer <LAYER>` : Aligns the cross-section with the voxels of
                                  the designated layer, and uses the radiological
                                  convention.
"""  # noqa: E122, E501
)

_clihelp.layout = textwrap.dedent(
f"""
Change the viewer's layout (i.e., the quadrants and their layers)

Neuroglancer has 8 different window types:
. xy     : {R("X")}{G("Y")} cross-section
. yz     : {G("Y")}{B("Z")} cross-section
. xz     : {R("X")}{B("Z")} cross-section
. xy-3d  : {R("X")}{G("Y")} cross-section in a {b("3D")} window
. yz-3d  : {G("Y")}{B("Z")} cross-section in a {b("3D")} window
. xz-3d  : {R("X")}{B("Z")} cross-section in a {b("3D")} window
. 4panel : Four quadrants ({R("X")}{G("Y")}, {R("X")}{B("Z")}, {b("3D")}, {G("Y")}{B("Z")})
. 3d     : {b("3D")} window

It is possible to build a user-defined layout by stacking these basic
windows into a row or a column -- or even nested rows and columns --
using the {b("--stack")} option. The {b("--layer")} option allows assigning
specific layers to a specific window. We also define {b("--append")} and
{b("--insert")} to add a new window into an existing stack of windows.
"""  # noqa: E122, E501
)

_clihelp.channel_mode = textwrap.dedent(
f"""
In neuroglancer, axes can be interpreted in three different ways:

. global  : The axis is common to all layers and can be navigated
            (default for axes that have spatial or temporal units)

. local   : The axis is specific to a layers and navigation is local
            (dimension names end with ')

. channel : The axis is specific to a layer; enables multi-channel shading
            (dimension names end with ^)

{b("Notes")}
{b("-----")}

. When an axis is local ('), only one channel can be shown at once.

. When an axis is a channel (^), all channels can be mixed into a single
  view using a multi-channel shader.

. When an axis is global, navigation is linked across layers.
"""  # noqa: E122, E501
)



from nibabel.streamlines.trk import TrkFile
import nibabel as nib
import numpy as np
import fsspec
import math
import random
import json


TRK = "https://dandiarchive.s3.amazonaws.com/blobs/d4a/c43/d4ac43bd-6896-4adf-a911-82edbea21f67"


with fsspec.open(TRK) as f:
    trk = TrkFile.load(f)


def make_sharding_info(nb_chunks: int) -> dict:
    preshift_bits = 12
    minishard_bits = int(math.ceil(math.log2(math.ceil(
        nb_chunks / 2**preshift_bits
    ))))
    return {
        "@type": "neuroglancer_uint64_sharded_v1",
        "hash": "identity",
        "preshift_bits": preshift_bits,
        "minishard_bits": minishard_bits,
        "shard_bits": 0,
        "minishard_index_encoding": "raw",
        "data_encoding": "raw",
    }


def make_info(verts, edges, scalars, spatial: list[dict]) -> dict:


    nb_streamlines = len(verts)
    lb = np.min(np.stack([np.min(x, axis=0) for x in verts]), axis=0)
    ub = np.max(np.stack([np.max(x, axis=0) for x in verts]), axis=0)

    return {
        "@type": "neuroglancer_annotations_v1",
        "dimensions": {"x": [1, "mm"], "y": [1, "mm"], "z": [1, "mm"]},
        "lower_bound": lb.tolist(),
        "upper_bound": ub.tolist(),
        "annotation_type": "LINE",
        "properties": [
            {
                "id": "orientation_color",
                "type": "rgb",
                "description": "Color-coding of the segment orientation",
            },
            {
                "id": "orientation_x",
                "type": "float32",
                "description": "Segment orientation (x)",
            },
            {
                "id": "orientation_y",
                "type": "float32",
                "description": "Segment orientation (y)",
            },
            {
                "id": "orientation_z",
                "type": "float32",
                "description": "Segment orientation (z)",
            },
            *[
                {
                    "id": key,
                    "type": "float32",
                }
                for key in scalars.keys()
            ]
        ],
        "relationships": [
            {
                "id": "tract",
                "key": "by_tract",
                "sharding": make_sharding_info(len(edges)),
            },
        ],
        "by_id": {
            "key": "by_id",
            "sharding": make_sharding_info(sum(map(len, edges))),
        },
        "spatial": spatial,
        # [
        #     # coarsest level
        #     {
        #     "key": "0",
        #     "sharding": { ... },
        #     "grid_shape": [1, 1, 1],
        #     "chunk_size": upper_bound - lower_bound,
        #     "limit": compute_limit(0, ...),

        #     },
        #     # finer level
        #     {
        #     "key": f"{nb_levels}",
        #     "sharding": { ... },
        #     "grid_shape": [ ... ],
        #     "chunk_size": [ ... ],
        #     "limit": compute_limit(nb_levels, ...),
        #     },
        # ],
        }


def make_spatial(edges, orient, scalars, mode="local") -> list[dict]:
    grid = [1, 1, 1]
    while True:



def shuffle(x: list) -> list:
    """Deterministic shuffling (makes a copy)"""
    x = list(x)
    random.seed(1234)
    random.shuffle(x)
    return x


def to_annot(inp: str, out: str, mode="local"):

    with fsspec.open(inp) as f:
        trk = TrkFile.load(f)

    verts = shuffle(trk.streamlines)
    scalars = {
        key: shuffle(val)
        for key, val in trk.tractogram.data_per_point.items()
    }

    edges = [np.stack([x[:-1:2], x[1::2]]) for x in verts]
    orient = [x[:, 1, :] - x[:, 0, :] for x in verts]
    orient = [x / (x*x).sum(-1, keepaxis=True)**0.5 for x in verts]
    scalars = {
        k: [(x[:-1:2] + x[1::2]) / 2 for x in v]
        for k, v in zip(scalars.items())
    }

    spatial_grids, spatial_info = make_spatial(edges, orient, scalars, mode)

    info = make_info(verts, edges, scalars, spatial_info)

    with open(f"{out}/info", "wb") as f:
        json.dump(info, f)

    with open(f"{out}/by_id/0.shard", "wb") as f:
        f.write(dump_by_id(verts, orient, scalars), f)

    for i, grid in enumerate(spatial_grids):
        with open(f"{out}/{i}/0.shard", "wb") as f:
            f.write(dump_spatial(grid, verts, orient, scalars), f)



"""
Tools to convert between units.

Functions
---------
split_unit
    Split prefix and unit
convert_unit
    Convert a value between units
normalize_unit
    Convert to prefixless SI unit
unit_to_scale
    Return the scale of a arbitrary unit with respect to its SI equivalent
as_ome_unit
    Find OME version of a unit
as_nifti_unit
    Find NIfTI version of a unit
as_neuroglancer_unit
    Find Neuroglancer version of a unit (= short SI)

Attributes
----------
UnitDomain
    Type hint for unit domains, _i.e._, `Literal["space", "time", "all"]`
OME_UNITS : dict[UnitDomain, list[str]]
    Dictionary where values are lists of valid OME units.
NIFTI_UNITS : dict[UnitDomain, list[str]]
    Dictionary where values are lists of valid NIfTI units.
SI_UNITS_SHORT : dict[UnitDomain, list[str]]
    Dictionary where values are lists of valid Neuroglancer (short SI) units.
SI_UNITS_LONG : dict[UnitDomain, list[str]]
    Dictionary where values are lists of valid SI (long) units.
SI_PREFIX_SHORT2LONG : dict[str, str]
    Dictionary that maps short forms of SI prefixes to long forms
SI_PREFIX_LONG2SHORT : dict[str, str]
    Dictionary that maps long forms of SI prefixes to short forms
SI_PREFIX_EXPONENT : dict[str, float]
    Dictionary that maps short forms of SI prefixes to their exponent value
UNIT_SHORT2LONG : dict[UnitDomain, dict[str, str]]
    Dictionary where values are dictionaries that map short forms of a
    unit to long forms
UNIT_LONG2SHORT : dict[UnitDomain, dict[str, str]]
    Dictionary where values are dictionaries that map long forms of a
    unit to short forms
"""
# stdlib
from typing import Literal

# externals
from numpy.typing import ArrayLike

_MU1 = '\u00B5'
_MU2 = '\u03BC'

UnitDomain = Literal["space", "time", "frequency", "angular_frequency", "all"]
UNIT_DOMAINS = UnitDomain.__args__

# ======================================================================
#   OME-NGFF
# ======================================================================

OME_UNITS: dict[UnitDomain, list[str]] = {
    "space": [
        "angstrom",
        "attometer",
        "centimeter",
        "decimeter",
        "exameter",
        "femtometer",
        "foot",
        "gigameter",
        "hectometer",
        "inch",
        "kilometer",
        "megameter",
        "meter",
        "micrometer",
        "mile",
        "millimeter",
        "nanometer",
        "parsec",
        "petameter",
        "picometer",
        "terameter",
        "yard",
        "yoctometer",
        "yottameter",
        "zeptometer",
        "zettameter",
    ],
    "time": [
        "attosecond",
        "centisecond",
        "day",
        "decisecond",
        "exasecond",
        "femtosecond",
        "gigasecond",
        "hectosecond",
        "hour",
        "kilosecond",
        "megasecond",
        "microsecond",
        "millisecond",
        "minute",
        "nanosecond",
        "petasecond",
        "picosecond",
        "second",
        "terasecond",
        "yoctosecond",
        "yottasecond",
        "zeptosecond",
        "zettasecond",
    ]
}
OME_UNITS["all"] = [unit for units in OME_UNITS.values() for unit in units]

# ======================================================================
#   NIFTI
# ======================================================================

NIFTI_UNITS: dict[UnitDomain, list[str]] = {
    "space": [
        "meter",
        "mm",
        "micron",
    ],
    "time": [
        "sec",
        "msec",
        "usec",
    ],
    "frequency": [
        "hz",
    ],
    "angular_frequency": [
        "rads",
        "rad/s",
    ]
}
NIFTI_UNITS["all"] = [unit for units in OME_UNITS.values() for unit in units]
NIFTI_UNITS["all"] += ["unknown", "ppm"]

# ======================================================================
#   SI PREFIXES
# ======================================================================

SI_PREFIX_SHORT2LONG: dict[str, str] = {
    "Q": "quetta",
    "R": "ronna",
    "Y": "yotta",
    "Z": "zetta",
    "E": "exa",
    "P": "peta",
    "T": "tera",
    "G": "giga",
    "M": "mega",
    "K": "kilo",
    "k": "kilo",
    "H": "hecto",
    "h": "hecto",
    "D": "deca",
    "da": "deca",
    "": "",
    "d": "deci",
    "c": "centi",
    "m": "milli",
    "u": "micro",
    _MU1: "micro",
    _MU2: "micro",
    "n": "nano",
    "p": "pico",
    "f": "femto",
    "a": "atto",
    "z": "zepto",
    "y": "yocto",
    "r": "ronto",
    "q": "quecto",
}

SI_PREFIX_LONG2SHORT: dict[str, str] = {
    long: short
    for short, long in SI_PREFIX_SHORT2LONG.items()
}

SI_PREFIX_EXPONENT: dict[str, int] = {
    "Q": 30,
    "R": 27,
    "Y": 24,
    "Z": 21,
    "E": 18,
    "P": 15,
    "T": 12,
    "G": 9,
    "M": 6,
    "K": 3,
    "k": 3,
    "H": 2,
    "h": 2,
    "D": 1,
    "da": 1,
    "": 0,
    "d": -1,
    "c": -2,
    "m": -3,
    "u": -6,
    _MU1: -6,
    _MU2: -6,
    "n": -9,
    "p": -12,
    "f": -15,
    "a": -18,
    "z": -21,
    "y": -24,
    "r": -27,
    "q": -30,
}

# ======================================================================
#   SI/NEUROGLANCER UNITS
# ======================================================================

UNIT_SHORT2LONG: dict[UnitDomain, dict[str, str]] = {}
"""Mapping from short name to long name (per domain)."""

UNIT_LONG2SHORT: dict[UnitDomain, dict[str, str]] = {}
"""Mapping from long name to short name (per domain)."""

UNIT_SCALE: dict[UnitDomain, dict[str, float]] = {}
"""Mapping from short name to scale (per domain)."""

SI_UNITS_SHORT: dict[UnitDomain, list[str]] = {}
"""List of short SI units (per category)."""

SI_UNITS_LONG: dict[UnitDomain, list[str]] = {}
"""List of long SI units (per domain)."""

# ----------------------------------------------------------------------
#   SPACE
# ----------------------------------------------------------------------

# SI
UNIT_SHORT2LONG["space"] = {
    short + "m": long + "meter"
    for short, long in SI_PREFIX_SHORT2LONG.items()
}
UNIT_SCALE["space"] = {
    prefix + "m": 10**exponent
    for prefix, exponent in SI_PREFIX_EXPONENT.items()
}

SI_UNITS_SHORT["space"] = list(UNIT_SHORT2LONG["space"].keys())
SI_UNITS_LONG["space"] = list(UNIT_SHORT2LONG["space"].values())

# Non-SI
UNIT_SHORT2LONG["space"].update({
    "mi": "mile",
    "yd": "yard",
    "ft": "foot",
    "in": "inch",
    """: "foot",
    """: "inch",
    "Å": "angstrom",
    "pc": "parsec",
})
UNIT_SCALE["space"].update({
    "mi": 1609.344,
    "yd": 0.9144,
    "ft": 0.3048,
    "'": 0.3048,
    "in": 25.4E-3,
    '"': 25.4E-3,
    "Å": 1E-10,
    "pc": 3.0857E16,
})

# Inverse map
UNIT_LONG2SHORT["space"] = {
    long: short
    for short, long in UNIT_SHORT2LONG["space"].items()
}
UNIT_LONG2SHORT["space"]["micron"] = "u"

# ----------------------------------------------------------------------
#   TIME
# ----------------------------------------------------------------------

# SI
UNIT_SHORT2LONG["time"] = {
    short + "s": long + "second"
    for short, long in SI_PREFIX_SHORT2LONG.items()
}
UNIT_SCALE["time"] = {
    prefix + "s": 10**exponent
    for prefix, exponent in SI_PREFIX_EXPONENT.items()
}

SI_UNITS_SHORT["time"] = list(UNIT_SHORT2LONG["time"].keys())
SI_UNITS_LONG["time"] = list(UNIT_SHORT2LONG["time"].values())

# Non-SI
UNIT_SHORT2LONG["time"].update({
    "y": "year",
    "d": "day",
    "h": "hour",
    "m": "minute",
    "s": "second",
})
UNIT_SCALE["time"].update({
    "y": 365.25*24*60*60,
    "d": 24*60*60,
    "h": 60*60,
    "m": 60,
})

# Inverse map
UNIT_LONG2SHORT["time"] = {
    long: short
    for short, long in UNIT_SHORT2LONG["time"].items()
}


# ----------------------------------------------------------------------
#   FREQUENCY
# ----------------------------------------------------------------------

# Only Hertz is used in neuroglancer and nifti so we only deal with
# this single unit.
# TODO: use a proper unit library to handle conversions?

# SI
UNIT_SHORT2LONG["frequency"] = {
    short + "hz": long + "hertz"
    for short, long in SI_PREFIX_SHORT2LONG.items()
}
UNIT_SHORT2LONG["frequency"].update({
    short + "Hz": long + "hertz"
    for short, long in SI_PREFIX_SHORT2LONG.items()
})
UNIT_SCALE["frequency"] = {
    prefix + "hz": 10**exponent
    for prefix, exponent in SI_PREFIX_EXPONENT.items()
}

SI_UNITS_SHORT["frequency"] = list(UNIT_SHORT2LONG["frequency"].keys())
SI_UNITS_LONG["frequency"] = list(UNIT_SHORT2LONG["frequency"].values())

# Inverse map
UNIT_LONG2SHORT["frequency"] = {
    long: short
    for short, long in UNIT_SHORT2LONG["frequency"].items()
}

# ----------------------------------------------------------------------
#   ANGULAR FREQUENCY
# ----------------------------------------------------------------------

# Only rad/s is used in neuroglancer and nifti so we only deal with
# this single unit.
# TODO: use a proper unit library to handle conversions?

# SI
UNIT_SHORT2LONG["angular_frequency"] = {
    short + "rad/s": long + "radian/second"
    for short, long in SI_PREFIX_SHORT2LONG.items()
}
UNIT_SHORT2LONG["angular_frequency"].update({
    short + "rads": long + "radian/second"
    for short, long in SI_PREFIX_SHORT2LONG.items()
})
UNIT_SCALE["angular_frequency"] = {
    prefix + "rad/s": 10**exponent
    for prefix, exponent in SI_PREFIX_EXPONENT.items()
}

SI_UNITS_SHORT["angular_frequency"] \
    = list(UNIT_SHORT2LONG["angular_frequency"].keys())
SI_UNITS_LONG["angular_frequency"] \
    = list(UNIT_SHORT2LONG["angular_frequency"].values())

# Inverse map
UNIT_LONG2SHORT["angular_frequency"] = {
    long: short
    for short, long in UNIT_SHORT2LONG["angular_frequency"].items()
}


# ----------------------------------------------------------------------
#   ALL
# ----------------------------------------------------------------------


SI_UNITS_SHORT["all"] = [u for units in SI_UNITS_SHORT.values() for u in units]
SI_UNITS_LONG["all"] = [u for units in SI_UNITS_LONG.values() for u in units]


# ----------------------------------------------------------------------
#   PUBLIC FUNCTIONS
# ----------------------------------------------------------------------


def split_unit(unit: str) -> tuple[str, str]:
    """
    Split prefix and unit.

    Parameters
    ----------
    unit : [sequence of] str

    Returns
    -------
    prefix : [sequence of] str
    unit : [sequence of] str
    """
    unit = as_neuroglancer_unit(unit)

    if unit == "":
        return ("", "")

    if isinstance(unit, (list, tuple)):
        unit = type(unit)(map(split_unit, unit))
        prefix = type(unit)(map(lambda x: x[0], unit))
        unit = type(unit)(map(lambda x: x[1], unit))
        return prefix, unit

    for strippedunit in ("rad/s", "hz", "m", "s"):
        if unit.endswith(strippedunit):
            return unit[:-len(strippedunit)], strippedunit
    if not unit:
        return '', ''
    raise ValueError(f'Unknown unit "{unit}"')


def same_unit_kind(src: str, dst: str) -> bool:
    """Check whether two units are of the same kind."""
    src = as_neuroglancer_unit(src)
    dst = as_neuroglancer_unit(dst)
    if src == "" and dst == "":
        return True
    if src == "" or dst == "":
        return False
    return split_unit(src)[1] == split_unit(dst)[1]


def convert_unit(
    value: ArrayLike | float,
    src: str,
    dst: str,
) -> ArrayLike | float:
    """Convert a value between different units.

    Parameters
    ----------
    value : float or ArrayLike
        Value to convert
    src : str
        Source unit
    dst : str
        Destination unit

    Returns
    -------
    value : float or array
        Converted value
    """
    if isinstance(value, (list, tuple)):
        src = ensure_list(src, len(value))
        dst = ensure_list(dst, len(value))
        return type(value)([
            convert_unit(x, s, d)
            for x, s, d in zip(value, src, dst)
        ])
    src = unit_to_scale(src)
    dst = unit_to_scale(dst)
    return value * (src / dst)


def normalize_unit(
    value: ArrayLike | float,
    unit: str
) -> ArrayLike | float:
    """
    Normalize a unit (i.e., convert to prefixless SI unit).

    Parameters
    ----------
    value : float or ArrayLike
        Value to convert
    unit : str
        Source unit

    Returns
    -------
    value : float or array
        Converted value
    """
    kind = split_unit(unit)[1]
    return convert_unit(value, unit, kind)


def ensure_list(x: object, n: int | None = None, **kwargs: dict) -> list:
    """
    Ensure that `x` is a list.

    Arrays are converted to nested lists, whereas scalars are insereted
    into a list of length 1.

    Parameters
    ----------
    x : object
        Input value(s)
    n : int, optional
        Target length of the list.
        If the input is longer, it gets cropped.
        Otherwise, it gets padded with the default value.
    default : any, optional
        Value to use to pad the list.
        If not provided, use last value in the list.

    Returns
    -------
    x : list
    """
    if hasattr(x, 'tolist'):
        x = x.tolist()
    if not isinstance(x, (list, tuple)):
        x = [x]
    x = list(x)
    if n is not None:
        if len(x) < n:
            default = kwargs.get('default', x[-1])
            x += [default] * (n - len(x))
        elif len(x) > n:
            x = x[:n]
    return x


def as_short_unit(unit: str) -> str:
    """
    Return the shortest representation for this unit.

    Parameters
    ----------
    unit : str
        Any unit.

    Returns
    -------
    unit : str
        An short unit.

    Raises
    ------
    ValueError
        If the input unit is unknown.

    """
    if unit == "":
        return unit
    if unit.endswith("sec"):
        # nifti
        unit = unit[:-2]
    for units in UNIT_SHORT2LONG.values():
        if unit in units:
            return unit
    for units in UNIT_LONG2SHORT.values():
        if unit in units:
            return units[unit]
    if unit in SI_PREFIX_SHORT2LONG:
        return unit
    if unit in SI_PREFIX_LONG2SHORT:
        return SI_PREFIX_LONG2SHORT[unit]
    raise ValueError("Unknown unit:", unit)


def as_long_unit(unit: str) -> str:
    """
    Return the long representation for this unit.

    Parameters
    ----------
    unit : str
        Any unit.

    Returns
    -------
    unit : str
        An long unit.

    Raises
    ------
    ValueError
        If the input unit is unknown.

    """
    if unit == "":
        return unit
    if unit.endswith("sec"):
        # nifti
        unit = unit[:-2]
    for units in UNIT_LONG2SHORT.values():
        if unit in units:
            return unit
    for units in UNIT_SHORT2LONG.values():
        if unit in units:
            return units[unit]
    if unit in SI_PREFIX_LONG2SHORT:
        return unit
    if unit in SI_PREFIX_SHORT2LONG:
        return SI_PREFIX_SHORT2LONG[unit]
    raise ValueError("Unknown unit:", unit)


def as_neuroglancer_unit(unit: str) -> str:
    """Return the Neuroglancer representation of a unit.

    Parameters
    ----------
    unit : str
        Any unit.

    Returns
    -------
    unit : str
        An OME-compatible unit.

    Raises
    ------
    ValueError
        If the input unit does not have an exactly equivalent
        representation in OME.
    """
    if unit == "":
        return unit
    short_unit = as_short_unit(unit)
    if short_unit not in SI_UNITS_SHORT["all"]:
        raise ValueError("Input is not a SI unit:", unit)
    # Couple of fixes for non-neuroglancer aliases
    if short_unit.endswith("Hz"):
        short_unit = short_unit[:-2] + "hz"
    if short_unit.endswith("rads"):
        short_unit = short_unit[:-4] + "rad/s"
    if short_unit[:1] in (_MU1, _MU2):
        short_unit = "u" + short_unit[1:]
    return short_unit


def as_ome_unit(unit: str) -> str:
    """Return the OME-compatible representation of a unit.

    Parameters
    ----------
    unit : str
        Any unit

    Returns
    -------
    unit : str
        An OME-compatible unit

    Raises
    ------
    ValueError
        If the input unit does not have an exactly equivalent
        representation in OME.
    """
    if unit == "":
        return unit
    long_unit = as_long_unit(unit)
    if long_unit not in OME_UNITS["all"]:
        raise ValueError("Input is not a OME unit:", unit)
    return long_unit


def as_nifti_unit(unit: str) -> str:
    """
    Return the NIfTI-compatible representation of a unit.

    Parameters
    ----------
    unit : str
        Any unit

    Returns
    -------
    unit : str
        A NIfTI-compatible unit.
        If the unit does not have a NIfTI representation, the NIfTI type
        `"unknown"` is returned.
    """
    if unit == "":
        return "unknown"
    unit = as_ome_unit(unit)
    return {
        "meter": "meter",
        "millimeter": "mm",
        "micrometer": "micron",
        "second": "sec",
        "millisecond": "msec",
        "microsecond": "usec",
    }.get(unit, "unknown")


def unit_to_scale(unit: str) -> float:
    """
    Return the scale of a arbitrary unit with respect to its SI equivalent.

    Parameters
    ----------
    unit : str
        Any unit

    Returns
    -------
    scale : float
        The corresponding SI scale

    Raises
    ------
    ValueError
        If the unit does not have a SI equivalent.
    """
    if unit == "":
        return 1.0
    if unit in UNIT_LONG2SHORT["space"]:
        unit = UNIT_LONG2SHORT["space"][unit]
    elif unit in UNIT_LONG2SHORT["time"]:
        unit = UNIT_LONG2SHORT["time"][unit]
    elif unit in SI_PREFIX_LONG2SHORT:
        unit = SI_PREFIX_LONG2SHORT[unit]
    if unit in UNIT_SCALE["space"]:
        unit = UNIT_SCALE["space"][unit]
    elif unit in UNIT_SCALE["time"]:
        unit = UNIT_SCALE["time"][unit]
    elif unit in SI_PREFIX_EXPONENT:
        unit = 10 ** SI_PREFIX_EXPONENT[unit]
    if isinstance(unit, str):
        raise ValueError("Unknown unit", unit)
    return unit




"""
An ArgumentParser that can be used as a commandline app.
It handles history and autocomplete.

We use it to implement a shell-like UI for neuroglancer.
"""
# stdlib
import argparse
import logging
import os
import re
import readline  # autocomplete/history in user input
import shlex  # parse user input as if a shell commandline
import sys
import traceback
import warnings
from functools import partial
from gettext import gettext  # to fix usage string

# internals
from ngtools.local.iostream import StandardIO
from ngtools.local.termcolors import bformat, iformat

LOG = logging.getLogger(__name__)


class Console(argparse.ArgumentParser):
    """
    An ArgumentParser that can be used as a commandline app.
    It handles history and autocomplete.
    """

    DEFAULT_HISTFILE = os.path.expanduser(
        os.environ.get("NGTOOLS_HISTFILE", '~/.neuroglancer_history')
    )
    DEFAULT_HISTSIZE = int(os.environ.get("NGTOOLS_HISTSIZE", 1000))

    def __init__(self, *args: tuple, **kwargs: dict) -> None:
        """
        Parameters
        ----------
        prog : str, default=os.path.basename(sys.argv[0])
            The name of the program.
        usage : str, optional
            A usage message. Default: auto-generated from arguments.
        description : str, default=""
            A description of what the program does.
        epilog : str, default=""
            Text following the argument descriptions
        prefix_chars : str, default="-"
            Characters that prefix optional arguments
        fromfile_prefix_chars: str | None, default=None
            Characters that prefix files containing additional arguments
        argument_default : object, default=None
            The default value for all arguments
        conflict_handler : {"error", "resolve"}, default="error"
            String indicating how to handle conflicts
        add_help : bool, default=True
            Add a -h/-help option
        allow_abbrev: bool, default=True
            Allow long options to be abbreviated unambiguously

        Other Parameters
        ----------------
        debug : bool, default=False
            Print traceback on error.
        history_file : str, default="~/.neuroglancer_history"
            Path to history file.
        history_size : int, default=1000
            History size.
        exit_on_error : bool, default=False
            Exit (non-gracefully) on error
        exit_on_help : bool, default=False
            Exist on help.
        stdin : TextIO | str, default=sys.stdin
            Input stream.
        stdout : TextIO | str, default=sys.stdout
            Output stream.
        stderr : TextIO | str, default=sys.stderr
            Error stream.
        max_choices : int | None, default=None
            Maximum numer of choices to show in usage string.
        """
        self._debug = kwargs.pop('debug', False)
        max_choices = kwargs.pop('max_choices', None)
        # Commandline behavior
        self.history_file = kwargs.pop('history_file', self.DEFAULT_HISTFILE)
        if self.history_file:
            self.history_file = os.path.expanduser(self.history_file)
        self.history_size = kwargs.pop('history_size', self.DEFAULT_HISTSIZE)
        # Exit behavior
        exit_on_error = kwargs.pop('exit_on_error', False)
        exit_on_help = kwargs.pop('exit_on_help', False)
        kwargs["exit_on_error"] = False
        # Input/output
        stdio = kwargs.pop("stdio", None)
        if stdio is None:
            stdin = kwargs.pop("stdin", sys.stdin)
            stdout = kwargs.pop("stdout", sys.stdout)
            stderr = kwargs.pop("stderr", sys.stderr)
            level = kwargs.pop("level", "info")
            if self._debug:
                level = "debug"
            stdio = StandardIO(
                stdin=stdin, stdout=stdout, stderr=stderr,
                level=level, logger=LOG
            )
        self.stdio = stdio
        # ArgumentParser.__init__
        kwargs["formatter_class"] = partial(
            MainHelpFormatter, max_choices=max_choices
        )
        super().__init__(*args, **kwargs)
        # Overwrite ArgumentParser attributes
        self.exit_on_error = exit_on_error
        self.exit_on_help = exit_on_help

    class InterruptParsing(Exception):
        """Exception raised when parsing gets interrupted."""

    def exit(self, status: int = 0, message: str | None = None) -> None:
        """Overload ArgumentParser.exit to disable it."""
        # We overload (and do not call) ArgumentParser.exit
        # This is called (e.g.) after running a help command.
        if message:
            self.print(message)
        raise self.InterruptParsing(status)

    def error(self, *args, **kwargs) -> None:
        """Error -- Print something to the stderr in red."""
        # We overload (and do not call) ArgumentParser.error
        if self.exit_on_error:
            raise SystemExit(*args)
        else:
            self.stdio.error(*args, **kwargs)

    def print(self, *args, **kwargs) -> None:  # noqa: D102
        self.stdio.print(*args, **kwargs)

    def debug(self, *args, **kwargs) -> None:  # noqa: D102
        self.stdio.debug(*args, **kwargs)

    def info(self, *args, **kwargs) -> None:  # noqa: D102
        self.stdio.info(*args, **kwargs)

    def warning(self, *args, **kwargs) -> None:  # noqa: D102
        self.stdio.warning(*args, **kwargs)

    def input(self, *args, **kwargs) -> str:  # noqa: D102
        return self.stdio.input(*args, **kwargs)

    def enter_console(self) -> None:
        """Set up history and auto-complete."""
        # NOTE key bindings
        #   ^[A : arrow up
        #   ^[B : arrow down
        #   ^[C : arrow right
        #   ^[D : arrow left
        # https://www.gnu.org/software/bash/manual/html_node/Commands-For-History.html
        readline.set_completer_delims(' \t\n;')
        readline.parse_and_bind("tab: complete")
        readline.parse_and_bind(r'"\e[A": previous-history')
        readline.parse_and_bind(r'"\e[B": next-history')
        readline.set_completer(self.complete)
        if self.history_file and self.history_size:
            if not os.path.exists(self.history_file):
                with open(self.history_file, 'wt'):
                    pass
            try:
                readline.read_history_file(self.history_file)
            except OSError:
                # https://github.com/neuroscales/ngtools/issues/16
                # -> Let's raise a warning -- better than crashing
                warnings.warn("Could not read command line history")
            readline.set_history_length(self.history_size)

    def exit_console(self) -> None:
        """Save history."""
        if self.history_file and self.history_size:
            readline.set_history_length(self.history_size)
            readline.write_history_file(self.history_file)

    def await_input(self) -> None:
        """Wait for next user input."""
        self.enter_console()

        self.print(
            f'\nType {bformat.bold("help")} to list available '
            f'commands, or {bformat.bold("help <command>")} '
            f'for specific help.\n'
            f'Type {bformat.bold("Ctrl+C")} to interrupt the '
            f'current command and {bformat.bold("Ctrl+D")} to '
            f'exit the app.'
        )

        def are_you_sure() -> bool:
            try:
                # Query input
                args = self.input("\nDo you really want to exit ([y]/n)?")
                args.strip()
                if args[:1].lower() == "n":
                    return False
                else:
                    raise EOFError
            except KeyboardInterrupt:
                # Ctrl+C -> do not exit
                return False

        count = 1
        try:
            while True:
                try:
                    # Query input
                    args = self.input(iformat.fg.green(f'[{count}] '))
                    args = args.strip()
                    if not args:
                        continue
                    if args[:1] == "#":
                        # Allow for "comments"
                        continue
                    count += 1
                except EOFError:
                    # Ctrl+D -> propagate and catch later
                    if not are_you_sure():
                        continue
                except KeyboardInterrupt:
                    # Ctrl+C -> generate new input
                    self.print("")
                    continue

                try:
                    # Parse
                    args = self.parse_args(shlex.split(args))
                    if not vars(args):
                        raise ValueError("Unknown command")
                except EOFError:
                    # Ctrl+D -> propagate and catch later
                    if not are_you_sure():
                        continue
                except self.InterruptParsing:
                    # Caught "exit" call in parser. Silent it.
                    # (This is triggered by --help)
                    continue
                except KeyboardInterrupt:
                    # Ctrl+C -> Stop parsing
                    count += 1
                    continue
                except Exception as e:
                    # Other exceptions -> print + new input field
                    if self._debug:
                        self.debug(traceback.print_tb(e.__traceback__))
                    self.error("(PARSE ERROR)", e)
                    continue

                try:
                    # Execute
                    func = getattr(args, 'func', lambda x: None)
                    func(args)
                except EOFError:
                    # Ctrl+D -> propagate and catch later
                    if not are_you_sure():
                        continue
                except self.InterruptParsing:
                    # Caught "exit" call in parser. Silent it.
                    # (This is triggered by --help)
                    continue
                except (Exception, KeyboardInterrupt) as e:
                    # Other exceptions -> print + new input field
                    if self._debug:
                        self.debug(traceback.print_tb(e.__traceback__))
                    self.error(f"(EXEC ERROR) [{type(e).__name__}]", e)
                    continue

        except EOFError:
            # Ctrl+D -> graceful exit
            self.print('exit')
            raise SystemExit
        finally:
            self.exit_console()

    @property
    def parsers(self) -> argparse._SubParsersAction | None:
        """Return registered sub parsers."""
        parsers = None
        for action in self._actions:
            if isinstance(action, argparse._SubParsersAction):
                parsers = action
                break
        return parsers

    @property
    def subcommands(self) -> list[str]:
        """Return the names of all subcommands."""
        if getattr(self.parsers, 'choices', None):
            return list(self.parsers.choices.keys())
        return []

    def _listdir(self, root: str) -> list[str]:
        """List directory 'root' appending the path separator to subdirs."""
        res = []
        for name in os.listdir(root):
            path = os.path.join(root, name)
            if os.path.isdir(path):
                name += os.sep
            res.append(name)
        return res

    def _complete_path(self, path: str | None = None) -> str:
        """Perform completion of filesystem path."""
        if not path:
            return self._listdir('.')
        dirname, rest = os.path.split(path)
        tmp = dirname if dirname else '.'
        res = [os.path.join(dirname, p)
               for p in self._listdir(tmp) if p.startswith(rest)]
        # more than one match, or single match which does not exist (typo)
        if len(res) > 1 or not os.path.exists(path):
            return res
        # resolved to a single directory, so return list of files below it
        if os.path.isdir(path):
            return [os.path.join(path, p) for p in self._listdir(path)]
        # exact file match terminates this completion
        return [path + ' ']

    def complete_default(self, context: str) -> str:
        """Fallback autocompleter."""
        return self._complete_path(os.path.expanduser(context))

    RE_SPACE = re.compile(r'.*\s+$', re.M)

    def complete(self, context: str, state: str) -> str:
        """Entry point for `readline` completion."""
        line = readline.get_line_buffer()
        begidx, endidx = readline.get_begidx(), readline.get_endidx()
        args = shlex.split(line)

        # show matching commands
        if not args or begidx <= len(args[0]):
            addspace = len(line) <= endidx or line[endidx] != ' '
            addspace = ' ' if addspace else ''
            result = [c + addspace for c in self.subcommands
                      if c.startswith(context)]

        # resolve command to the implementation function
        else:
            cmd = args[0].strip()
            if cmd in self.subcommands:
                template = 'complete_' + cmd
                impl = getattr(self, template, self.complete_default)
            else:
                impl = self.complete_default
            result = impl(context)

        try:
            return result[state]
        except IndexError:
            return None


class FixOrderHelpFormatter(argparse.HelpFormatter):
    """
    Fix help to say that positionals must be given before options.

    argparse default usage string says that positional arguments
    should be given after optional arguments, whereas it's really the
    opposite.

    The ordering happens in the middle of the long-ish _format_usage
    method, so I see no other fix than copying and fixing the full
    method.
    """

    def _format_usage(  # noqa: ANN202
        self: argparse.HelpFormatter,
        usage: str | None,
        actions: list[argparse.Action],
        groups: list[argparse._ArgumentGroup],
        prefix: str | None,
    ) -> str:
        if prefix is None:
            prefix = gettext('usage: ')

        # if usage is specified, use that
        if usage is not None:
            usage = usage % dict(prog=self._prog)

        # if no optionals or positionals are available, usage is just prog
        elif usage is None and not actions:
            usage = '%(prog)s' % dict(prog=self._prog)

        # if optionals and positionals are available, calculate usage
        elif usage is None:
            prog = '%(prog)s' % dict(prog=self._prog)

            # split optionals from positionals
            optionals = []
            positionals = []
            for action in actions:
                if action.option_strings:
                    optionals.append(action)
                else:
                    positionals.append(action)

            # build full usage string
            format = self._format_actions_usage
            action_usage = format(positionals + optionals, groups)
            usage = ' '.join([s for s in [prog, action_usage] if s])

            # wrap the usage parts if it's too long
            text_width = self._width - self._current_indent
            if len(prefix) + len(usage) > text_width:

                # break usage into wrappable parts
                part_regexp = (
                    r'\(.*?\)+(?=\s|$)|'
                    r'\[.*?\]+(?=\s|$)|'
                    r'\S+'
                )
                opt_usage = format(optionals, groups)
                pos_usage = format(positionals, groups)
                opt_parts = re.findall(part_regexp, opt_usage)
                pos_parts = re.findall(part_regexp, pos_usage)
                assert ' '.join(opt_parts) == opt_usage
                assert ' '.join(pos_parts) == pos_usage

                # helper for wrapping lines
                def get_lines(
                        parts: list[str],
                        indent: str,
                        prefix: str | None = None
                ) -> list[str]:
                    lines = []
                    line = []
                    indent_length = len(indent)
                    if prefix is not None:
                        line_len = len(prefix) - 1
                    else:
                        line_len = indent_length - 1
                    for part in parts:
                        if line_len + 1 + len(part) > text_width and line:
                            lines.append(indent + ' '.join(line))
                            line = []
                            line_len = indent_length - 1
                        line.append(part)
                        line_len += len(part) + 1
                    if line:
                        lines.append(indent + ' '.join(line))
                    if prefix is not None:
                        lines[0] = lines[0][indent_length:]
                    return lines

                # if prog is short, follow it with optionals or positionals
                if len(prefix) + len(prog) <= 0.75 * text_width:
                    indent = ' ' * (len(prefix) + len(prog) + 1)
                    if pos_parts:
                        lines = get_lines([prog] + pos_parts, indent, prefix)
                    elif opt_parts:
                        lines = get_lines([prog] + opt_parts, indent, prefix)
                        lines.extend(get_lines(pos_parts, indent))
                    else:
                        lines = [prog]

                # if prog is long, put it on its own line
                else:
                    indent = ' ' * len(prefix)
                    parts = pos_parts + opt_parts
                    lines = get_lines(parts, indent)
                    if len(lines) > 1:
                        lines = []
                        lines.extend(get_lines(pos_parts, indent))
                        lines.extend(get_lines(opt_parts, indent))
                    lines = [prog] + lines

                # join lines into usage
                usage = '\n'.join(lines)

        # prefix with 'usage:'
        return '%s%s\n\n' % (prefix, usage)


class MaxChoiceHelpFormatter(argparse.HelpFormatter):
    """Only display up to `max_choices` choices in usage string."""

    def _metavar_formatter(
        self, action, default_metavar  # noqa: ANN001
    ) -> callable:
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            # --- begin patch ---
            choices = list(action.choices)
            max_choices = getattr(self, 'max_choices', None)
            if max_choices is not None and len(choices) > max_choices:
                choices = choices[:max_choices] + ['...']
            result = '{%s}' % ','.join(map(str, choices))
            # ---  end patch  ---
        else:
            result = default_metavar

        def format(tuple_size: int) -> tuple:
            if isinstance(result, tuple):
                return result
            else:
                return (result, ) * tuple_size
        return format


class ActionHelpFormatter(
    FixOrderHelpFormatter,
    argparse.RawTextHelpFormatter,
):
    """Formatter used for commands."""


class MainHelpFormatter(
    FixOrderHelpFormatter,
    MaxChoiceHelpFormatter,
    argparse.RawTextHelpFormatter,
):
    """Formatter used for the main console."""

    def __init__(self, *args, **kwargs) -> None:
        self.max_choices = kwargs.pop("max_choices", None)
        super().__init__(*args, **kwargs)



"""Read/write AFNI's transforms."""
from math import pi
import numpy as np
from nibabel.affines import (
    obliquity,
    voxel_sizes,
)

from .base import (
    BaseLinearTransformList,
    DisplacementsField,
    LinearParameters,
    TransformFileError,
    _ensure_image,
)

LPS = np.diag([-1, -1, 1, 1])
OBLIQUITY_THRESHOLD_DEG = 0.01


class AFNILinearTransform(LinearParameters):
    """A string-based structure for AFNI linear transforms."""

    def __str__(self):
        """Generate a string representation."""
        param = self.structarr["parameters"]
        return "\t".join(["%g" % p for p in param[:3, :].reshape(-1)])

    def to_string(self, banner=True):
        """Convert to a string directly writeable to file."""
        string = "%s\n" % self
        if banner:
            string = "\n".join(
                ("# 3dvolreg matrices (DICOM-to-DICOM, row-by-row):", string)
            )
        return string % self

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an AFNI affine from a nitransform's RAS+ matrix.

        AFNI implicitly de-obliques image affine matrices before applying transforms, so
        for consistency we update the transform to account for the obliquity of the images.

        .. testsetup:
            >>> import pytest
            >>> pytest.skip()

        >>> moving.affine == ras @ reference.affine

        We can decompose the affines into oblique and de-obliqued components:

        >>> moving.affine == m_obl @ m_deobl
        >>> reference.affine == r_obl @ r_deobl

        To generate an equivalent AFNI transform, we need an effective transform (``e_ras``):

        >>> m_obl @ m_deobl == ras @ r_obl @ r_deobl
        >>> m_deobl == inv(m_obl) @ ras @ r_obl @ r_deobl

        Hence,

        >>> m_deobl == e_ras @ r_deobl
        >>> e_ras == inv(m_obl) @ ras @ r_obl
        """
        # swapaxes is necessary, as axis 0 encodes series of transforms

        reference = _ensure_image(reference)
        if reference is not None and _is_oblique(reference.affine):
            ras = ras @ _cardinal_rotation(reference.affine, False)

        moving = _ensure_image(moving)
        if moving is not None and _is_oblique(moving.affine):
            ras = _cardinal_rotation(moving.affine, True) @ ras

        # AFNI represents affine transformations as LPS-to-LPS
        parameters = np.swapaxes(LPS @ ras @ LPS, 0, 1)

        tf = cls()
        tf.structarr["parameters"] = parameters.T
        return tf

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        tf = cls()
        sa = tf.structarr
        lines = [
            line
            for line in string.splitlines()
            if line.strip()
            and not (line.startswith("#") or "3dvolreg matrices" in line)
        ]

        if not lines:
            raise TransformFileError

        try:
            parameters = np.vstack(
                (
                    np.genfromtxt([lines[0].encode()], dtype="f8").reshape((3, 4)),
                    (0.0, 0.0, 0.0, 1.0),
                )
            )
        except ValueError as e:
            raise TransformFileError from e

        sa["parameters"] = parameters
        return tf

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms internal RAS+ matrix."""
        # swapaxes is necessary, as axis 0 encodes series of transforms
        retval = LPS @ np.swapaxes(self.structarr["parameters"].T, 0, 1) @ LPS
        reference = _ensure_image(reference)
        if reference is not None and _is_oblique(reference.affine):
            retval = retval @ _cardinal_rotation(reference.affine, True)

        moving = _ensure_image(moving)
        if moving is not None and _is_oblique(moving.affine):
            retval = _cardinal_rotation(moving.affine, False) @ retval

        return retval


class AFNILinearTransformArray(BaseLinearTransformList):
    """A string-based structure for series of AFNI linear transforms."""

    _inner_type = AFNILinearTransform

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms' internal RAS matrix."""

        pre_rotation = post_rotation = np.eye(4)
        if reference is not None and _is_oblique(ref_aff := _ensure_image(reference).affine):
            pre_rotation = _cardinal_rotation(ref_aff, True)
        if moving is not None and _is_oblique(mov_aff := _ensure_image(moving).affine):
            post_rotation = _cardinal_rotation(mov_aff, False)

        return np.stack([
            post_rotation @ (xfm.to_ras() @ pre_rotation)
            for xfm in self.xforms
        ])

    def to_string(self):
        """Convert to a string directly writeable to file."""
        strings = []
        for i, xfm in enumerate(self.xforms):
            lines = [
                line.strip()
                for line in xfm.to_string(banner=(i == 0)).splitlines()
                if line.strip()
            ]
            strings += lines
        return "\n".join(strings + [""])

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an ITK affine from a nitransform's RAS+ matrix."""
        _self = cls()

        pre_rotation = post_rotation = np.eye(4)

        if reference is not None and _is_oblique(ref_aff := _ensure_image(reference).affine):
            pre_rotation = _cardinal_rotation(ref_aff, False)
        if moving is not None and _is_oblique(mov_aff := _ensure_image(moving).affine):
            post_rotation = _cardinal_rotation(mov_aff, True)

        _self.xforms = [
            cls._inner_type.from_ras(post_rotation @ ras[i, ...] @ pre_rotation)
            for i in range(ras.shape[0])
        ]
        return _self

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        _self = cls()

        lines = [
            line.strip()
            for line in string.splitlines()
            if line.strip() and not line.startswith("#")
        ]
        if not lines:
            raise TransformFileError("Input string is empty.")

        _self.xforms = [cls._inner_type.from_string(line) for line in lines]
        return _self


class AFNIDisplacementsField(DisplacementsField):
    """A data structure representing displacements fields."""

    @classmethod
    def from_image(cls, imgobj):
        """Import a displacements field from a NIfTI file."""
        hdr = imgobj.header.copy()
        shape = hdr.get_data_shape()

        if len(shape) != 5 or shape[-2] != 1 or not shape[-1] in (2, 3):
            raise TransformFileError(
                'Displacements field "%s" does not come from AFNI.'
                % imgobj.file_map["image"].filename
            )

        field = np.squeeze(np.asanyarray(imgobj.dataobj))
        field[..., (0, 1)] *= -1.0

        return imgobj.__class__(field, imgobj.affine, hdr)

    @classmethod
    def to_image(cls, imgobj):
        """Export a displacements field from a nibabel object."""

        hdr = imgobj.header.copy()

        warp_data = imgobj.get_fdata().reshape(imgobj.shape[:3] + (1, imgobj.shape[-1]))
        warp_data[..., (0, 1)] *= -1

        return imgobj.__class__(warp_data, imgobj.affine, hdr)


def _is_oblique(affine, thres=OBLIQUITY_THRESHOLD_DEG):
    """
    Determine whether the dataset is oblique.

    Examples
    --------
    >>> _is_oblique(np.eye(4))
    False

    >>> _is_oblique(nb.affines.from_matvec(
    ...     nb.eulerangles.euler2mat(x=0.9, y=0.001, z=0.001),
    ...     [4.0, 2.0, -1.0],
    ... ))
    True

    """
    return float(obliquity(affine).max() * 180 / pi) > thres


def _afni_deobliqued_grid(oblique, shape):
    """
    Calculate AFNI's target deobliqued image grid.

    Maps the eight images corners to the new coordinate system to ensure
    coverage of the full extent after rotation, as AFNI does.

    See also
    --------
    https://github.com/afni/afni/blob/75766463758e5806d938c8dd3bdcd4d56ab5a485/src/mri_warp3D.c#L941-L1010

    Parameters
    ----------
    oblique : 4x4 numpy.array
        affine that is not aligned to the cardinal axes.
    shape : numpy.array
        sizes of the (oblique) image grid

    Returns
    -------
    affine : 4x4 numpy.array
        plumb affine (i.e., aligned to the cardinal axes).
    shape : numpy.array
        sizes of the target, plumb image grid

    """
    shape = np.array(shape[:3])
    vs = voxel_sizes(oblique)

    # Calculate new shape of deobliqued grid
    corners_ijk = (
        np.array(
            [
                (i, j, k)
                for k in (0, shape[2])
                for j in (0, shape[1])
                for i in (0, shape[0])
            ]
        )
        - 0.5
    )
    corners_xyz = oblique @ np.hstack((corners_ijk, np.ones((len(corners_ijk), 1)))).T
    extent = corners_xyz.min(1)[:3], corners_xyz.max(1)[:3]
    nshape = ((extent[1] - extent[0]) / vs + 0.999).astype(int)

    # AFNI deobliqued target will be in LPS+ orientation
    plumb = LPS * ([vs.min()] * 3 + [1.0])

    # Coordinates of center voxel do not change
    obliq_c = oblique @ np.hstack((0.5 * (shape - 1), 1.0))
    plumb_c = plumb @ np.hstack((0.5 * (nshape - 1), 1.0))

    # Rebase the origin of the new, plumb affine
    plumb[:3, 3] -= plumb_c[:3] - obliq_c[:3]

    return plumb, nshape


def _dicom_real_to_card(oblique):
    """
    Calculate the corresponding "DICOM cardinal" for "DICOM real" (AFNI jargon).

    Implements the internal "deobliquing" operation of ``3drefit`` and other tools, which
    just *drop* the obliquity from the input affine.

    Parameters
    ----------
    oblique : 4x4 numpy.array
        affine that may not be aligned to the cardinal axes ("IJK_DICOM_REAL" for AFNI).

    Returns
    -------
    plumb : 4x4 numpy.array
        affine aligned to the cardinal axes ("IJK_DICOM_CARD" for AFNI).

    """
    # Origin is kept from input
    retval = np.eye(4)
    retval[:3, 3] = oblique[:3, 3]

    # Calculate director cosines and project to closest canonical
    cosines = oblique[:3, :3] / np.abs(oblique[:3, :3]).max(0)
    cosines[np.abs(cosines) < 1.0] = 0
    # Once director cosines are calculated, scale by voxel sizes
    retval[:3, :3] = np.round(voxel_sizes(oblique), decimals=4) * cosines
    return retval


def _cardinal_rotation(oblique, real_to_card=True):
    """
    Calculate the rotation matrix to undo AFNI's deoblique operation.

    Parameters
    ----------
    oblique : 4x4 numpy.array
        affine that may not be aligned to the cardinal axes ("IJK_DICOM_REAL" for AFNI).

    Returns
    -------
    plumb : 4x4 numpy.array
        affine aligned to the cardinal axes ("IJK_DICOM_CARD" for AFNI).

    """
    card = _dicom_real_to_card(oblique)
    return (
        card @ np.linalg.inv(oblique) if real_to_card else oblique @ np.linalg.inv(card)
    )


def _afni_warpdrive(oblique, forward=True):
    """
    Calculate AFNI's ``WARPDRIVE_MATVEC_FOR_000000`` (de)obliquing affine.

    Parameters
    ----------
    oblique : 4x4 numpy.array
        affine that is not aligned to the cardinal axes.
    forward : :obj:`bool`
        Returns the forward transformation if True, i.e.,
        the matrix to convert an oblique affine into an AFNI's plumb (if ``True``)
        or viceversa plumb -> oblique (if ``false``).

    Returns
    -------
    warpdrive : 4x4 numpy.array
        AFNI's *warpdrive* forward or inverse matrix.

    """
    ijk_to_dicom_real = np.diag(LPS) * oblique
    ijk_to_dicom = _dicom_real_to_card(oblique)
    R = np.linalg.inv(ijk_to_dicom) @ ijk_to_dicom_real
    return np.linalg.inv(R) if forward else R


def _afni_header(nii, field="WARPDRIVE_MATVEC_FOR_000000", to_ras=False):
    from lxml import etree

    root = etree.fromstring(nii.header.extensions[0].get_content().decode())
    retval = np.fromstring(
        root.find(f".//*[@atr_name='{field}']").text, sep="\n", dtype="float32"
    )
    if retval.size == 12:
        retval = np.vstack((retval.reshape((3, 4)), (0, 0, 0, 1)))
        if to_ras:
            retval = LPS @ retval @ LPS

    return retval



"""Tests for the `spaces` module."""
from ngtools import spaces as S
from ngtools.optionals import try_import_as

ng = try_import_as('neuroglancer', 'ngtools.nglite')


def test_get_neuronames() -> None:
    """List of all possible neurological spaces."""
    ref1 = {"r", "l", "a", "p", "i", "s"}
    assert set(S._get_neuronames(1)) == ref1

    ref2 = {
        "ra", "rp", "la", "lp", "ar", "al", "pr", "pl",
        "ri", "rs", "li", "ls", "ir", "il", "sr", "sl",
        "ai", "as", "pi", "ps", "ia", "ip", "sa", "sp",
    }
    assert set(S._get_neuronames(2)) == ref2

    ref3 = {
        "rai", "ras", "rpi", "rps", "lai", "las", "lpi", "lps",
        "ria", "rip", "rsa", "rsp", "lia", "lip", "lsa", "lsp",
        "ari", "ars", "ali", "als", "pri", "prs", "pli", "pls",
        "air", "ail", "asr", "asl", "pir", "pil", "psr", "psl",
        "ira", "irp", "ila", "ilp", "sra", "srp", "sla", "slp",
        "iar", "ial", "ipr", "ipl", "sar", "sal", "spr", "spl",
    }
    assert set(S._get_neuronames(3)) == ref3


def test_get_defaultnames() -> None:
    """List of all possible default spaces."""
    ref1 = {"x", "y", "z"}
    assert set(S._get_defaultnames(1)) == ref1

    ref2 = {"xy", "yx", "xz", "zx", "yz", "zy"}
    assert set(S._get_defaultnames(2)) == ref2

    ref3 = {"xyz", "xzy", "yxz", "yzx", "zxy", "zyx"}
    assert set(S._get_defaultnames(3)) == ref3


def test_space_is_compatible() -> None:
    """Compatibility between different oriented spaces."""
    assert S.space_is_compatible("xyz", "ras")
    assert S.space_is_compatible("xyz", "sar")
    assert S.space_is_compatible("xyz", "lpi")
    assert S.space_is_compatible("ras", "lpi")

    assert S.space_is_compatible("xy", "ra")
    assert S.space_is_compatible("xy", "ar")
    assert S.space_is_compatible("yz", "pi")
    assert S.space_is_compatible("as", "pi")

    assert not S.space_is_compatible("xyz", "rac")
    assert not S.space_is_compatible("xyz", "ra")
    assert not S.space_is_compatible("ra", "pi")


def test_convert_space() -> None:
    """Test simple conversion."""
    # NOTE: we use `5 * 1E-4` and not `5E-4` because the conversion
    # code uses the former and both values differ in double precision
    # arithmetic.

    inp = ng.CoordinateSpace({
        "x": [5, "um"],
        "y": [5, "mm"],
        "z": [5, "cm"],
        "t": [4, "ms"],
        "c": [1, ""],
    })

    ref = ng.CoordinateSpace({
        "x": [5 * 1E-4, "cm"],
        "y": [5 * 1E-3, "m"],
        "z": [5 * 1E-5, "km"],
        "t": [4 * 1E-5, "hs"],
        "c": [1, ""],
    })

    out = S.convert_space(inp, {"x": "cm", "y": "m", "z": "km", "t": "hs"})
    assert out.to_json() == ref.to_json(), \
           f"Convert to arbitrary units:\nout={out}\nref={ref}"

    ref = ng.CoordinateSpace({
        "x": [5 * 1E-6, "m"],
        "y": [5 * 1E-3, "m"],
        "z": [5 * 1E-2, "m"],
        "t": [4 * 1E-3, "s"],
        "c": [1, ""],
    })

    out = S.normalize_space(inp)
    assert out.to_json() == ref.to_json(), \
           f"Convert to SI:\nout={out}\nref={ref}"



#!/usr/bin/env python
# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tool for creating videos (sequences of screenshots) from Neuroglancer.

The video is specified by a "script", which is a text file of the form:

  <neuroglancer-url>
  <transition-duration>
  <neuroglancer-url>
  ...
  <neuroglancer-url>
  <transition-duration>
  <neuroglancer-url>

The transition-duration is a floating point number encoded in text format that
specifies the length of the transition between the two states in seconds.  Each
transition corresponds to smoothly interpolating between the adjacent states
over the specified duration: positions are interpolated linearly, orientations
are interpolated using spherical linear interpolation, and zoom factors are
interpolated exponentially.

Note that combining zooming with position movement in a single transition (as
can occur when using the mouse wheel rather than the keyboard commands to zoom
out) may produce a poor interpolation result, because the position will appear
to move at a much faster rate when zoomed in than when zoomed out.

Using the `edit` command, you can interactively create or edit a script.

The `render` command is used to render an existing script to a sequence of PNG
images.  Frames are captured only once all data has been loaded; because of
this, rendering can be slow.

The layer panel and other UI widgets, as well as the borders around panels, are
disabled when rendering.  Additionally, the image size is determined by the
command line `--width` and `--height` arguments, rather than the size of the
browser window.  However, the states in the script determine whether the scale
bar and axis lines are shown.  Therefore, make sure to set their visibility
appropriately when creating the script.

The `toggle-play` command in the editor can be used to preview the transitions.
However, due to limitations in the Neuroglancer Python interface, the frame rate
is limited to 5 frames per second, and what is displayed differs from what will
be rendered in several ways:
  - it does not wait for all data to be loaded;
  - the UI controls and panel borders are shown;
  - the width and height are based on the browser size.
"""

import argparse
import bisect
import copy
import math
import os
import threading
import time
import webbrowser

import atomicwrites

import neuroglancer
import neuroglancer.cli
import neuroglancer.tool.screenshot


class PlaybackManager:
    def __init__(self, keypoints, frames_per_second):
        self.keypoints = keypoints
        self.frames_per_second = frames_per_second

        self.total_frames = 0
        self.keypoint_start_frame = []
        self.keypoint_end_frame = []
        for k in keypoints[:-1]:
            duration = k["transition_duration"]
            if duration == 0:
                cur_frames = 0
            else:
                cur_frames = max(
                    1, int(round(k["transition_duration"] * frames_per_second))
                )
            self.keypoint_start_frame.append(self.total_frames)
            self.total_frames += cur_frames
            self.keypoint_end_frame.append(self.total_frames)
        self.keypoint_start_frame.append(self.total_frames)
        self.total_frames += 1
        self.keypoint_end_frame.append(self.total_frames)

    def get_keypoint_from_frame(self, frame_i):
        if frame_i < 0 or frame_i >= self.total_frames:
            raise ValueError
        return bisect.bisect_right(self.keypoint_start_frame, frame_i) - 1

    def get_frame_from_elapsed_time(self, elapsed_time):
        return int(math.floor(elapsed_time * self.frames_per_second))

    def get_frame(self, frame_i):
        start_keypoint = self.get_keypoint_from_frame(frame_i)
        a = self.keypoints[start_keypoint]["state"]
        if start_keypoint == len(self.keypoints) - 1:
            return a
        else:
            end_keypoint = start_keypoint + 1
            b = self.keypoints[end_keypoint]["state"]
            start_frame = self.keypoint_start_frame[start_keypoint]
            end_frame = self.keypoint_end_frame[start_keypoint]
            t = (frame_i - start_frame) / (end_frame - start_frame)
            return neuroglancer.ViewerState.interpolate(a, b, t)

    def get_frames(self, start_frame, end_frame):
        return [
            self.get_frame(frame_i)
            for frame_i in range(start_frame, min(end_frame, self.total_frames))
        ]

    def set_state(self, viewer, frame_i, prefetch_frames):
        states = self.get_frames(frame_i, frame_i + prefetch_frames)
        viewer.set_state(states[0])
        with viewer.config_state.txn() as s:
            del s.prefetch[:]
            for i, state in enumerate(states[1:]):
                s.prefetch.append(
                    neuroglancer.PrefetchState(
                        state=state, priority=prefetch_frames - i
                    )
                )


class EditorPlaybackManager:
    def __init__(self, script_editor, playing=True, frames_per_second=5):
        self.script_editor = script_editor
        self.frames_per_second = frames_per_second
        self.playback_manager = PlaybackManager(
            script_editor.keypoints, frames_per_second=self.frames_per_second
        )
        self.current_keypoint_index = max(1, script_editor.keypoint_index)
        self.script_editor._set_keypoint_index(self.current_keypoint_index)
        self.playing = playing
        script_editor.playback_manager = self
        self.current_frame = self.playback_manager.keypoint_start_frame[
            self.current_keypoint_index - 1
        ]
        self.start_time = (
            time.time() - self.current_frame / self.playback_manager.frames_per_second
        )
        t = threading.Thread(target=self._thread_func)
        t.daemon = True
        t.start()
        self.should_stop = threading.Event()
        self._update()

    def _update_current_frame(self):
        elapsed_time = time.time() - self.start_time
        self.current_frame = self.playback_manager.get_frame_from_elapsed_time(
            elapsed_time
        )

    def _display_frame(self):
        frame_i = self.current_frame
        keypoint_index = (
            self.playback_manager.get_keypoint_from_frame(
                min(frame_i, self.playback_manager.total_frames - 1)
            )
            + 1
        )
        current_duration = self.script_editor.keypoints[keypoint_index - 1][
            "transition_duration"
        ]
        transition_time = (
            frame_i - self.playback_manager.keypoint_start_frame[keypoint_index - 1]
        ) / self.playback_manager.frames_per_second
        self.playback_status = "%s frame %d/%d transition %.1f/%g" % (
            "PLAYING" if self.playing else "PAUSED",
            frame_i,
            self.playback_manager.total_frames,
            transition_time,
            current_duration,
        )
        if keypoint_index != self.current_keypoint_index:
            self.script_editor._set_keypoint_index(keypoint_index)
            self.current_keypoint_index = keypoint_index
        if frame_i >= self.playback_manager.total_frames:
            self.script_editor.playback_manager = None
            self.script_editor._update_status()
            self.should_stop.set()
            return
        self.playback_manager.set_state(
            self.script_editor.viewer, frame_i, prefetch_frames=10
        )
        self.script_editor._update_status()

    def reload(self):
        self.playback_manager = PlaybackManager(
            self.script_editor.keypoints, frames_per_second=self.frames_per_second
        )
        self.current_keypoint_index = None
        self.seek_frame(0)

    def pause(self):
        if self.playing:
            self.seek_frame(0)
        else:
            self.start_time = (
                time.time()
                - self.current_frame / self.playback_manager.frames_per_second
            )
            self.playing = True

    def seek_frame(self, amount):
        if self.playing:
            self.playing = False
            self._update_current_frame()
        self.current_frame += amount
        self.current_frame = max(
            0, min(self.current_frame, self.playback_manager.total_frames - 1)
        )
        self._display_frame()

    def _thread_func(self):
        while True:
            time.sleep(0.2)
            self.script_editor.viewer.defer_callback(self._update)
            if self.should_stop.is_set():
                return

    def _update(self):
        if self.script_editor.playback_manager is not self:
            self.should_stop.set()
            return
        if not self.playing:
            return
        self._update_current_frame()
        self._display_frame()


def load_script(script_path, transition_duration=1):
    keypoints = []
    with open(script_path) as f:
        while True:
            url = f.readline()
            if not url:
                break
            line = f.readline()
            if not line:
                duration = transition_duration
            else:
                duration = float(line)
            keypoints.append(
                {"state": neuroglancer.parse_url(url), "transition_duration": duration}
            )
    return keypoints


def save_script(script_path, keypoints):
    temp_path = script_path + ".tmp"
    with open(temp_path, "w") as f:
        for x in keypoints:
            f.write(neuroglancer.to_url(x["state"]) + "\n")
            f.write(str(x["transition_duration"]) + "\n")
    if hasattr(os, "replace"):
        # Only available on Python3
        os.replace(temp_path, script_path)
    else:
        # Fails on Windows if script_path already exists
        os.rename(temp_path, script_path)


class ScriptEditor:
    def __init__(
        self,
        script_path,
        transition_duration,
        fullscreen_width,
        fullscreen_height,
        fullscreen_scale_bar_scale,
        frames_per_second,
    ):
        self.script_path = script_path
        self.viewer = neuroglancer.Viewer()
        self.frames_per_second = frames_per_second
        self.default_transition_duration = transition_duration
        self.fullscreen_width = fullscreen_width
        self.fullscreen_height = fullscreen_height
        self.fullscreen_scale_bar_scale = fullscreen_scale_bar_scale
        self.keypoint_index = 0
        if os.path.exists(script_path):
            self.keypoints = load_script(script_path, self.default_transition_duration)
        else:
            self.keypoints = []

        self.transition_duration = transition_duration
        self.viewer.shared_state.add_changed_callback(self._viewer_state_changed)
        self.quit_event = threading.Event()
        self.is_dirty = True
        self.is_fullscreen = False
        keybindings = [
            ("keyk", "add-keypoint"),
            ("bracketleft", "prev-keypoint"),
            ("bracketright", "next-keypoint"),
            ("backspace", "delete-keypoint"),
            ("shift+bracketleft", "decrease-duration"),
            ("shift+bracketright", "increase-duration"),
            ("home", "first-keypoint"),
            ("end", "last-keypoint"),
            ("keyq", "quit"),
            ("enter", "toggle-play"),
            ("keyf", "toggle-fullscreen"),
            ("keyj", "revert-script"),
            ("comma", "prev-frame"),
            ("period", "next-frame"),
        ]
        with self.viewer.config_state.txn() as s:
            for k, a in keybindings:
                s.input_event_bindings.viewer[k] = a
                s.input_event_bindings.slice_view[k] = a
                s.input_event_bindings.perspective_view[k] = a
        self._keybinding_message = " ".join("{}={}".format(*x) for x in keybindings)
        self.viewer.actions.add("add-keypoint", self._add_keypoint)
        self.viewer.actions.add("prev-keypoint", self._prev_keypoint)
        self.viewer.actions.add("next-keypoint", self._next_keypoint)
        self.viewer.actions.add("delete-keypoint", self._delete_keypoint)
        self.viewer.actions.add("increase-duration", self._increase_duration)
        self.viewer.actions.add("decrease-duration", self._decrease_duration)
        self.viewer.actions.add("first-keypoint", self._first_keypoint)
        self.viewer.actions.add("last-keypoint", self._last_keypoint)
        self.viewer.actions.add("quit", self._quit)
        self.viewer.actions.add("toggle-play", self._toggle_play)
        self.viewer.actions.add("toggle-fullscreen", self._toggle_fullscreen)
        self.viewer.actions.add("revert-script", self._revert_script)
        self.viewer.actions.add("next-frame", self._next_frame)
        self.viewer.actions.add("prev-frame", self._prev_frame)
        self.playback_manager = None
        self._set_keypoint_index(1)

    def _revert_script(self, s):
        if os.path.exists(self.script_path):
            self.keypoints = load_script(
                self.script_path, self.default_transition_duration
            )
            if self.playback_manager is not None:
                self.playback_manager.reload()
            else:
                self._set_keypoint_index(self.keypoint_index)

    def _toggle_fullscreen(self, s):
        self.is_fullscreen = not self.is_fullscreen
        with self.viewer.config_state.txn() as s:
            if self.is_fullscreen:
                s.show_ui_controls = False
                s.show_panel_borders = False
                s.viewer_size = [self.fullscreen_width, self.fullscreen_height]
                s.scale_bar_options.scale_factor = self.fullscreen_scale_bar_scale
            else:
                s.show_ui_controls = True
                s.show_panel_borders = True
                s.viewer_size = None
                s.scale_bar_options.scale_factor = 1

    def _next_frame(self, s):
        if self.playback_manager is None:
            EditorPlaybackManager(
                self, playing=False, frames_per_second=self.frames_per_second
            )
        self.playback_manager.seek_frame(1)

    def _prev_frame(self, s):
        if self.playback_manager is None:
            EditorPlaybackManager(
                self, playing=False, frames_per_second=self.frames_per_second
            )
        self.playback_manager.seek_frame(-1)

    def _add_keypoint(self, s):
        self.keypoints.insert(
            self.keypoint_index,
            {"state": s.viewer_state, "transition_duration": self.transition_duration},
        )
        self.keypoint_index += 1
        self.is_dirty = False
        self.save()
        self._update_status()

    def _toggle_play(self, s):
        if self.playback_manager is not None:
            self.playback_manager.pause()
        else:
            EditorPlaybackManager(self, playing=True)

    def _stop_playback(self):
        self.playback_manager = None

    def _set_transition_duration(self, value):
        self._stop_playback()
        self.transition_duration = value
        if self.keypoint_index > 0:
            self.keypoints[self.keypoint_index - 1]["transition_duration"] = value
        self.save()
        self._update_status()

    def save(self):
        save_script(self.script_path, self.keypoints)

    def _increase_duration(self, s):
        self._set_transition_duration(self.transition_duration + 0.1)

    def _decrease_duration(self, s):
        self._set_transition_duration(self.transition_duration - 0.1)

    def _get_is_dirty(self):
        if self.keypoint_index == 0:
            return True
        state = self.keypoints[self.keypoint_index - 1]["state"]
        return state.to_json() != self.viewer.state.to_json()

    def _viewer_state_changed(self):
        if self.playback_manager is not None:
            return
        is_dirty = self._get_is_dirty()
        if is_dirty != self.is_dirty:
            self.is_dirty = is_dirty
            self._update_status()

    def _delete_keypoint(self, s):
        self._stop_playback()
        if self.keypoint_index > 0:
            del self.keypoints[self.keypoint_index - 1]
            self.save()
            self.keypoint_index = self.keypoint_index - 1
            self.is_dirty = self._get_is_dirty()
            self._update_status()

    def _set_keypoint_index(self, index):
        index = max(0, min(index, len(self.keypoints)))
        self.keypoint_index = index
        state_index = max(0, index - 1)
        if len(self.keypoints) > 0:
            self.viewer.set_state(self.keypoints[state_index]["state"])
            self.transition_duration = self.keypoints[state_index][
                "transition_duration"
            ]
            self.is_dirty = False
        else:
            self.is_dirty = True
        self._update_status()

    def _prev_keypoint(self, s):
        self._stop_playback()
        if self.is_dirty:
            self._set_keypoint_index(self.keypoint_index)
        else:
            self._set_keypoint_index(self.keypoint_index - 1)

    def _next_keypoint(self, s):
        self._stop_playback()
        self._set_keypoint_index(self.keypoint_index + 1)

    def _first_keypoint(self, s):
        self._stop_playback()
        self._set_keypoint_index(0)

    def _last_keypoint(self, s):
        self._stop_playback()
        self._set_keypoint_index(len(self.keypoints))

    def _update_status(self):
        if self.playback_manager is not None:
            dirty_message = self.playback_manager.playback_status
        elif self.is_dirty:
            dirty_message = " [ CHANGED ]"
        else:
            dirty_message = ""

        status = "[ Keypoint %d/%d ]%s [ transition duration %g s ]  %s" % (
            self.keypoint_index,
            len(self.keypoints),
            dirty_message,
            self.transition_duration,
            self._keybinding_message,
        )
        with self.viewer.config_state.txn() as s:
            s.status_messages["status"] = status

    def _quit(self, s):
        self.quit_event.set()


def run_edit(args):
    editor = ScriptEditor(
        script_path=args.script,
        transition_duration=args.duration,
        fullscreen_width=args.width,
        fullscreen_height=args.height,
        fullscreen_scale_bar_scale=args.scale_bar_scale,
        frames_per_second=args.fps,
    )
    print(editor.viewer)
    if args.browser:
        webbrowser.open_new(editor.viewer.get_viewer_url())
    editor.quit_event.wait()


def _get_states_to_capture(
    keypoints, fps, resume, output_directory
) -> list[tuple[int, float, neuroglancer.ViewerState, str]]:
    def get_output_path(frame_number: int) -> str:
        return os.path.join(args.output_directory, "%07d.png" % frame_number)

    states_to_capture: list[tuple[int, float, neuroglancer.ViewerState, str]] = []
    frame_number = 0
    for i in range(len(keypoints) - 1):
        a = keypoints[i]["state"]
        b = keypoints[i + 1]["state"]
        duration = keypoints[i]["transition_duration"]
        num_frames = max(1, int(duration * fps))
        for frame_i in range(num_frames):
            t = frame_i / num_frames
            path = get_output_path(frame_number)
            if not (resume and os.path.exists(path)):
                cur_state = neuroglancer.ViewerState.interpolate(a, b, t)
                states_to_capture.append((frame_number, i + t, cur_state, path))
            frame_number += 1
    return states_to_capture


def run_render(args):
    keypoints = load_script(args.script)
    for keypoint in keypoints:
        neuroglancer.tool.screenshot.apply_state_modifications(keypoint["state"], args)
    fps = args.fps
    total_frames = sum(max(1, k["transition_duration"] * fps) for k in keypoints[:-1])

    os.makedirs(args.output_directory, exist_ok=True)
    states_to_capture = _get_states_to_capture(
        keypoints, fps=fps, resume=args.resume, output_directory=args.output_directory
    )

    num_frames_written = [total_frames - len(states_to_capture)]

    lock = threading.Lock()

    def make_request(state_to_capture: tuple[int, int, neuroglancer.ViewerState, str]):
        frame_number, t, state, path = state_to_capture

        def config_callback(s):
            s.viewer_size = (args.width, args.height)
            s.scale_bar_options.scale_factor = args.scale_bar_scale

        frame_desc = "%.3f/%5d" % (t, len(keypoints))

        def response_callback(screenshot):
            with atomicwrites.atomic_write(path, mode="wb", overwrite=True) as f:
                f.write(screenshot.image)
            with lock:
                num_frames_written[0] += 1
                cur_num_frames_written = num_frames_written[0]
                print(
                    "[%07d/%07d] keypoint %s: %s"
                    % (cur_num_frames_written, total_frames, frame_desc, path)
                )

        return neuroglancer.tool.screenshot.CaptureScreenshotRequest(
            config_callback=config_callback,
            state=state,
            response_callback=response_callback,
            description=frame_desc,
        )

    with neuroglancer.tool.screenshot.get_viewers(args) as viewers:
        neuroglancer.tool.screenshot.capture_screenshots_in_parallel(
            viewers=viewers,
            request_iter=[
                make_request(state_to_capture) for state_to_capture in states_to_capture
            ],
            refresh_browser_timeout=args.refresh_browser_timeout,
            num_to_prefetch=args.prefetch,
            total_requests=len(states_to_capture),
        )


def run_pan(args):
    from scipy.spatial.transform import Rotation

    state = args.state

    duration = args.duration

    keypoints = []

    orig_orientation = state.projection_orientation
    for i in range(5):
        rotation = Rotation.from_rotvec([0, math.pi / 2 * i, 0])
        new_orientation = Rotation.from_quat(orig_orientation) * rotation
        new_state = copy.deepcopy(state)
        new_state.projection_orientation = new_orientation.as_quat()
        keypoints.append({"state": new_state, "transition_duration": duration / 4})

    save_script(args.script, keypoints)


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    neuroglancer.cli.add_server_arguments(ap)
    sub_aps = ap.add_subparsers(help="command to run")
    ap_edit = sub_aps.add_parser("edit", help="Create or edit a script.")
    ap_edit.set_defaults(func=run_edit)
    ap_render = sub_aps.add_parser("render", help="Render a script.")
    ap_render.set_defaults(func=run_render)
    ap_pan = sub_aps.add_parser("pan", help="Create 360 degree pan script.")
    ap_pan.set_defaults(func=run_pan)
    neuroglancer.cli.add_state_arguments(ap_pan, required=True)
    ap_pan.add_argument(
        "-d", "--duration", type=float, help="Pan duration in seconds.", default=4
    )

    for ap_sub in [ap_edit, ap_render, ap_pan]:
        ap_sub.add_argument("script", help="Path to script file to read and write.")

    ap_edit.add_argument(
        "-d",
        "--duration",
        type=float,
        help="Default transition duration in seconds.",
        default=1,
    )
    ap_edit.add_argument(
        "--browser", action="store_true", help="Open web browser automatically."
    )
    ap_edit.add_argument("--width", type=int, help="Frame width", default=1920)
    ap_edit.add_argument("--height", type=int, help="Frame height", default=1080)
    ap_edit.add_argument(
        "--scale-bar-scale", type=float, help="Scale factor for scale bar", default=1
    )
    ap_edit.add_argument(
        "-f", "--fps", type=float, help="Frames per second.", default=5
    )

    ap_render.add_argument(
        "output_directory", help="Directory in which to write screenshot frames."
    )
    ap_render.add_argument(
        "-f", "--fps", type=float, help="Frames per second.", default=24
    )
    ap_render.add_argument(
        "--resume", action="store_true", help="Skip already rendered frames."
    )
    neuroglancer.tool.screenshot.define_state_modification_args(ap_render)
    neuroglancer.tool.screenshot.define_viewer_args(ap_render)
    neuroglancer.tool.screenshot.define_size_args(ap_render)
    neuroglancer.tool.screenshot.define_capture_args(ap_render)

    args = ap.parse_args()
    neuroglancer.cli.handle_server_arguments(args)
    args.func(args)



"""Handling of standard IO + logging."""
import atexit
import logging
import sys
from io import StringIO
from os import PathLike
from typing import Literal, TextIO

from ngtools.local.termcolors import bformat

LogLevel = int | Literal["error", "warning", "info", "debug", "any"] | None
LOG = logging.getLogger(__name__)


class StandardIO:
    """An IO handling class."""

    def __init__(
        self,
        stdin: TextIO | PathLike | str = sys.stdin,
        stdout: TextIO | PathLike | str = sys.stdout,
        stderr: TextIO | PathLike | str = sys.stderr,
        level: LogLevel = "info",
        logger: logging.Logger = LOG,
    ) -> None:
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.logger = logger
        if isinstance(level, str):
            if level[0].lower() == "a":
                level = 0
            else:
                level = getattr(logging, level.upper())
        self.level = level
        atexit.register(self.__del__)

    @property
    def stdin(self) -> TextIO | PathLike | str:
        """Get input stream."""
        return self._stdin

    @stdin.setter
    def stdin(self, value: TextIO | PathLike | str) -> None:
        stdin_file = getattr(self, "_stdin_file", None)
        if stdin_file and not stdin_file.closed:
            stdin_file.close()
        self._stdin = value
        self._stdin_file = None

    def format(self, *args, **kwargs) -> str:
        """Format a message like `print` would."""
        with StringIO() as f:
            print(*args, **kwargs, file=f)
            msg = f.getvalue()
        return msg

    def print(self, *args, **kwargs) -> str:
        """Print something to the stdout (no logging)."""
        file = kwargs.pop("file", self.stdout)
        flush = kwargs.pop("flush", False)
        level = kwargs.pop("level", float('inf'))
        format = kwargs.pop("format", (lambda x: x))

        msg = self.format(*args, **kwargs)
        if file and self.level <= level:
            if isinstance(file, str):
                with open(file, "a") as f:
                    print(format(msg), file=f, flush=flush, end="")
            else:
                print(format(msg), file=file, flush=flush, end="")
        return msg

    def debug(self, *args, **kwargs) -> str:
        """Debug -- Print something to the stderr."""
        kwargs.setdefault("file", self.stderr)
        msg = self.print(*args, **kwargs, level=logging.DEBUG)
        if self.logger:
            self.logger.debug(msg)
        return msg

    def info(self, *args, **kwargs) -> str:
        """Info -- Print something to the stdout."""
        msg = self.print(*args, **kwargs, level=logging.INFO)
        if self.logger:
            self.logger.info(msg)
        return msg

    def warning(self, *args, **kwargs) -> str:
        """Warn -- Print something to the stdout in orange."""
        kwargs["format"] = bformat.warning
        msg = self.print(*args, **kwargs, level=logging.WARNING)
        if self.logger:
            self.logger.warning(msg)
        return msg

    def error(self, *args, **kwargs) -> str:
        """Warn -- Print something to the stderr in red."""
        kwargs["format"] = bformat.fail
        msg = self.print(*args, **kwargs, level=logging.ERROR)
        if self.logger:
            self.logger.error(msg)
        return msg

    def input(self, prompt: str = "") -> str:
        """Read input from stdin or from a file."""
        stdin = self.stdin
        if stdin is sys.stdin:
            return input(prompt)
        else:
            if not hasattr(stdin, "read"):
                stdin = getattr(self, "_stdin_file", None)
            if stdin is None:
                stdin = self._stdin_file = open(self.stdin)

            self.print(prompt, end="")
            while True:
                try:
                    inp = next(stdin).rstrip("\n").rstrip("\r").rstrip("\n")
                    self.print(inp)
                    return inp
                except StopIteration:
                    pass

    def __del__(self) -> None:
        if hasattr(self, "_stdin_file"):
            stdin = self._stdin_file
            if stdin and not stdin.closed:
                stdin.close()
        atexit.unregister(self.__del__)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Resampling utilities."""

import asyncio
from os import cpu_count
from contextlib import suppress
from functools import partial
from pathlib import Path
from typing import Callable, TypeVar, Union

import numpy as np
from nibabel.loadsave import load as _nbload
from nibabel.arrayproxy import get_obj_dtype
from nibabel.spatialimages import SpatialImage
from scipy import ndimage as ndi

from .base import (
    ImageGrid,
    TransformBase,
    TransformError,
    SpatialReference,
    _as_homogeneous,
)

R = TypeVar("R")

SERIALIZE_VOLUME_WINDOW_WIDTH: int = 8
"""Minimum number of volumes to automatically serialize 4D transforms."""


async def worker(job: Callable[[], R], semaphore) -> R:
    async with semaphore:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, job)


async def _apply_serial(
    data: np.ndarray,
    spatialimage: SpatialImage,
    targets: np.ndarray,
    transform: TransformBase,
    ref_ndim: int,
    ref_ndcoords: np.ndarray,
    n_resamplings: int,
    output: np.ndarray,
    input_dtype: np.dtype,
    order: int = 3,
    mode: str = "constant",
    cval: float = 0.0,
    prefilter: bool = True,
    max_concurrent: int = min(cpu_count(), 12),
):
    """
    Resample through a given transform serially, in a 3D+t setting.

    Parameters
    ----------
    data : :obj:`~numpy.ndarray`
        The input data array.
    spatialimage : :obj:`~nibabel.spatialimages.SpatialImage` or :obj:`os.PathLike`
        The image object containing the data to be resampled in reference
        space
    targets : :obj:`~numpy.ndarray`
        The target coordinates for mapping.
    transform : :obj:`~nitransforms.base.TransformBase`
        The 3D, 3D+t, or 4D transform through which data will be resampled.
    ref_ndim : :obj:`int`
        Dimensionality of the resampling target (reference image).
    ref_ndcoords : :obj:`~numpy.ndarray`
        Physical coordinates (RAS+) where data will be interpolated, if the resampling
        target is a grid, the scanner coordinates of all voxels.
    n_resamplings : :obj:`int`
        Total number of 3D resamplings (can be defined by the input image, the transform,
        or be matched, that is, same number of volumes in the input and number of transforms).
    output : :obj:`~numpy.ndarray`
        The output data array where resampled values will be stored volume-by-volume.
    order : :obj:`int`, optional
        The order of the spline interpolation, default is 3.
        The order has to be in the range 0-5.
    mode : :obj:`str`, optional
        Determines how the input image is extended when the resamplings overflows
        a border. One of ``'constant'``, ``'reflect'``, ``'nearest'``, ``'mirror'``,
        or ``'wrap'``. Default is ``'constant'``.
    cval : :obj:`float`, optional
        Constant value for ``mode='constant'``. Default is 0.0.
    prefilter: :obj:`bool`, optional
        Determines if the image's data array is prefiltered with
        a spline filter before interpolation. The default is ``True``,
        which will create a temporary *float64* array of filtered values
        if *order > 1*. If setting this to ``False``, the output will be
        slightly blurred if *order > 1*, unless the input is prefiltered,
        i.e. it is the result of calling the spline filter on the original
        input.

    Returns
    -------
    np.ndarray
        Data resampled on the 3D+t array of input coordinates.

    """
    tasks = []
    semaphore = asyncio.Semaphore(max_concurrent)

    for t in range(n_resamplings):
        xfm_t = (
            transform if (n_resamplings == 1 or transform.ndim < 4) else transform[t]
        )

        targets_t = (
            ImageGrid(spatialimage).index(
                _as_homogeneous(xfm_t.map(ref_ndcoords), dim=ref_ndim)
            )
            if targets is None
            else targets[t, ...]
        )

        data_t = (
            data
            if data is not None
            else spatialimage.dataobj[..., t].astype(input_dtype, copy=False)
        )

        tasks.append(
            asyncio.create_task(
                worker(
                    partial(
                        ndi.map_coordinates,
                        data_t,
                        targets_t,
                        output=output[..., t],
                        order=order,
                        mode=mode,
                        cval=cval,
                        prefilter=prefilter,
                    ),
                    semaphore,
                )
            )
        )
    await asyncio.gather(*tasks)
    return output


def apply(
    transform: TransformBase,
    spatialimage: Union[str, Path, SpatialImage],
    reference: Union[str, Path, SpatialImage] = None,
    order: int = 3,
    mode: str = "constant",
    cval: float = 0.0,
    prefilter: bool = True,
    output_dtype: np.dtype = None,
    dtype_width: int = 8,
    serialize_nvols: int = SERIALIZE_VOLUME_WINDOW_WIDTH,
    max_concurrent: int = min(cpu_count(), 12),
) -> Union[SpatialImage, np.ndarray]:
    """
    Apply a transformation to an image, resampling on the reference spatial object.

    Parameters
    ----------
    transform: :obj:`~nitransforms.base.TransformBase`
        The 3D, 3D+t, or 4D transform through which data will be resampled.
    spatialimage : :obj:`~nibabel.spatialimages.SpatialImage` or :obj:`os.PathLike`
        The image object containing the data to be resampled in reference
        space
    reference : :obj:`~nibabel.spatialimages.SpatialImage` or :obj:`os.PathLike`
        The image, surface, or combination thereof containing the coordinates
        of samples that will be sampled.
    order : :obj:`int`, optional
        The order of the spline interpolation, default is 3.
        The order has to be in the range 0-5.
    mode : :obj:`str`, optional
        Determines how the input image is extended when the resamplings overflows
        a border. One of ``'constant'``, ``'reflect'``, ``'nearest'``, ``'mirror'``,
        or ``'wrap'``. Default is ``'constant'``.
    cval : :obj:`float`, optional
        Constant value for ``mode='constant'``. Default is 0.0.
    prefilter : :obj:`bool`, optional
        Determines if the image's data array is prefiltered with
        a spline filter before interpolation. The default is ``True``,
        which will create a temporary *float64* array of filtered values
        if *order > 1*. If setting this to ``False``, the output will be
        slightly blurred if *order > 1*, unless the input is prefiltered,
        i.e. it is the result of calling the spline filter on the original
        input.
    output_dtype : :obj:`~numpy.dtype`, optional
        The dtype of the returned array or image, if specified.
        If ``None``, the default behavior is to use the effective dtype of
        the input image. If slope and/or intercept are defined, the effective
        dtype is float64, otherwise it is equivalent to the input image's
        ``get_data_dtype()`` (on-disk type).
        If ``reference`` is defined, then the return value is an image, with
        a data array of the effective dtype but with the on-disk dtype set to
        the input image's on-disk dtype.
    dtype_width : :obj:`int`
        Cap the width of the input data type to the given number of bytes.
        This argument is intended to work as a way to implement lower memory
        requirements in resampling.
    serialize_nvols : :obj:`int`
        Minimum number of volumes in a 3D+t (that is, a series of 3D transformations
        independent in time) to resample on a one-by-one basis.
        Serialized resampling can be executed concurrently (parallelized) with
        the argument ``max_concurrent``.
    max_concurrent : :obj:`int`
        Maximum number of 3D resamplings to be executed concurrently.

    Returns
    -------
    resampled : :obj:`~nibabel.spatialimages.SpatialImage` or :obj:`~numpy.ndarray`
        The data imaged after resampling to reference space.

    """
    if reference is not None and isinstance(reference, (str, Path)):
        reference = _nbload(str(reference))

    _ref = (
        transform.reference
        if reference is None
        else SpatialReference.factory(reference)
    )

    if _ref is None:
        raise TransformError("Cannot apply transform without reference")

    if isinstance(spatialimage, (str, Path)):
        spatialimage = _nbload(str(spatialimage))

    # Avoid opening the data array just yet
    input_dtype = cap_dtype(get_obj_dtype(spatialimage.dataobj), dtype_width)

    # Number of data volumes
    data_nvols = 1 if spatialimage.ndim < 4 else spatialimage.shape[-1]
    # Number of transforms: transforms chains (e.g., affine + field, are a single transform)
    xfm_nvols = 1 if transform.ndim < 4 else len(transform)

    if data_nvols != xfm_nvols and min(data_nvols, xfm_nvols) > 1:
        raise ValueError(
            "The fourth dimension of the data does not match the transform's shape."
        )

    serialize_nvols = (
        serialize_nvols if serialize_nvols and serialize_nvols > 1 else np.inf
    )
    n_resamplings = max(data_nvols, xfm_nvols)
    serialize_4d = n_resamplings >= serialize_nvols

    targets = None
    ref_ndcoords = _ref.ndcoords.T
    if hasattr(transform, "to_field") and callable(transform.to_field):
        targets = ImageGrid(spatialimage).index(
            _as_homogeneous(
                transform.to_field(reference=reference).map(ref_ndcoords),
                dim=_ref.ndim,
            )
        )
    else:
        # Targets' shape is (Nt, 3, Nv) with Nv = Num. voxels, Nt = Num. timepoints.
        targets = (
            ImageGrid(spatialimage).index(
                _as_homogeneous(transform.map(ref_ndcoords), dim=_ref.ndim)
            )
            if targets is None
            else targets
        )

    if targets.ndim == 3:
        targets = np.rollaxis(targets, targets.ndim - 1, 0)
    else:
        assert targets.ndim == 2
        targets = targets[np.newaxis, ...]

    if serialize_4d:
        data = (
            np.asanyarray(spatialimage.dataobj, dtype=input_dtype)
            if data_nvols == 1
            else None
        )

        # Order F ensures individual volumes are contiguous in memory
        # Also matches NIfTI, making final save more efficient
        resampled = np.zeros(
            (len(ref_ndcoords), n_resamplings), dtype=input_dtype, order="F"
        )

        resampled = asyncio.run(
            _apply_serial(
                data,
                spatialimage,
                targets,
                transform,
                _ref.ndim,
                ref_ndcoords,
                n_resamplings,
                resampled,
                input_dtype,
                order=order,
                mode=mode,
                cval=cval,
                prefilter=prefilter,
                max_concurrent=max_concurrent,
            )
        )
    else:
        data = np.asanyarray(spatialimage.dataobj, dtype=input_dtype)

        if data_nvols == 1 and xfm_nvols == 1:
            targets = np.squeeze(targets)
            assert targets.ndim == 2
        # Cast 3D data into 4D if 4D nonsequential transform
        elif data_nvols == 1 and xfm_nvols > 1:
            data = data[..., np.newaxis]

        if xfm_nvols > 1:
            assert targets.ndim == 3
            n_time, n_dim, n_vox = targets.shape
            # Reshape to (3, n_time x n_vox)
            ijk_targets = np.rollaxis(targets, 0, 2).reshape((n_dim, -1))
            time_row = np.repeat(np.arange(n_time), n_vox)[None, :]

            # Now targets is (4, n_vox x n_time), with indexes (t, i, j, k)
            # t is the slowest-changing axis, so we put it first
            targets = np.vstack((time_row, ijk_targets))
            data = np.rollaxis(data, data.ndim - 1, 0)

        resampled = ndi.map_coordinates(
            data,
            targets,
            order=order,
            mode=mode,
            cval=cval,
            prefilter=prefilter,
        )

    if isinstance(_ref, ImageGrid):  # If reference is grid, reshape
        hdr = (
            _ref.header.copy()
            if _ref.header is not None
            else spatialimage.header.__class__()
        )
        hdr.set_data_dtype(output_dtype or spatialimage.header.get_data_dtype())

        if serialize_4d:
            resampled = resampled.reshape(
                _ref.shape
                if n_resamplings == 1
                else _ref.shape + (resampled.shape[-1],)
            )
        else:
            resampled = resampled.reshape((-1, *_ref.shape))
            resampled = np.rollaxis(resampled, 0, resampled.ndim)
            with suppress(ValueError):
                resampled = np.squeeze(resampled, axis=3)

        moved = spatialimage.__class__(resampled, _ref.affine, hdr)
        return moved

    output_dtype = output_dtype or input_dtype
    return resampled.astype(output_dtype)


def cap_dtype(dt, nbytes):
    """
    Cap the datatype size to shave off memory requirements.

    Examples
    --------
    >>> cap_dtype(np.dtype('f8'), 4)
    dtype('float32')

    >>> cap_dtype(np.dtype('f8'), 16)
    dtype('float64')

    >>> cap_dtype('float64', 4)
    dtype('float32')

    >>> cap_dtype(np.dtype('i1'), 4)
    dtype('int8')

    >>> cap_dtype('int8', 4)
    dtype('int8')

    >>> cap_dtype('int32', 1)
    dtype('int8')

    >>> cap_dtype(np.dtype('i8'), 4)
    dtype('int32')

    """
    dt = np.dtype(dt)
    return np.dtype(f"{dt.byteorder}{dt.kind}{min(nbytes, dt.itemsize)}")



# stdlib
import math as pymath
from ast import literal_eval

# externals
import numpy as np

# internals
from .utils import ensure_list

affine_subbasis_choices = ('T', 'R', 'Z', 'Z0', 'I', 'S', 'SC')
affine_subbasis_aliases = dict(
    T=['translation', 'translations'],
    R=['rot', 'rotation', 'rotations'],
    Z=['zoom', 'zooms', 'scale', 'scales', 'scaling', 'scalings'],
    Z0=['isovolume', 'isovolumic'],
    I=['isotropic', 'isozoom', 'isoscale', 'isoscaling'],
    S=['shear', 'shears'],
)


def affine_subbasis(mode, dim=3, sub=None, dtype=None):
    """Generate a basis set for the algebra of some (Lie) group of matrices.

    The basis is returned in homogeneous coordinates, even if
    the group required does not require translations. To extract the
    linear part of the basis: lin = basis[:-1, :-1].

    This function focuses on very simple (and coherent) groups.

    Note that shears generated by the 'S' basis are not exactly the same
    as classical shears ('SC'). Setting one classical shear parameter to
    a non-zero value generally applies a gradient of translations along
    a direction:
    + -- +         + -- +
    |    |   =>   /    /
    + -- +       + -- +
    Setting one Lie shear parameter to a non zero value is more alike
    to performing an expansion in one (diagonal) direction and a
    contraction in the orthogonal (diagonal) direction. It is a bit
    harder to draw in ascii, but it can also be seen as a horizontal
    shear followed by a vertical shear.
    The 'S' basis is orthogonal to the 'R' basis, but the 'SC' basis is
    not.

    Parameters
    ----------
    mode : {'T', 'R', 'Z', 'Z0', 'I', 'S', 'SC'}
        Group that should be encoded by the basis set:
            * 'T'   : Translations                     [dim]
            * 'R'   : Rotations                        [dim*(dim-1)//2]
            * 'Z'   : Zooms (= anisotropic scalings)   [dim]
            * 'Z0'  : Isovolumic scalings              [dim-1]
            * 'I'   : Isotropic scalings               [1]
            * 'S'   : Shears (symmetric)               [dim*(dim-1)//2]
            * 'SC'  : Shears (classic)                 [dim*(dim-1)//2]
        If the group name is appended with a list of integers, they
        have the same use as ``sub``. For example 'R[0]' returns the
        first rotation basis only. This grammar cannot be used in
        conjunction with the ``sub`` keyword.

    dim : int, default=3
        Dimension

    sub : int or list[int], optional
        Request only subcomponents of the basis.

    dtype : torch.type, optional
        Data type of the returned array.


    Returns
    -------
    basis : (F, dim+1, dim+1) tensor[dtype]
        Basis set, where ``F`` is the number of basis functions.

    """

    # Authors
    # -------
    # .. John Ashburner <j.ashburner@ucl.ac.uk> : original Matlab code
    # .. Yael Balbastre <yael.balbastre@gmail.com> : Python code

    # Check if sub passed in mode
    mode = mode.split('[')
    if len(mode) > 1:
        if sub is not None:
            raise ValueError('Cannot use both ``mode`` and ``sub`` '
                             'to specify a sub-basis.')
        sub = '[' + mode[1]
        sub = literal_eval(sub)  # Safe eval for list of native types
    mode = mode[0]

    if mode not in affine_subbasis_choices:
        for key, aliases in affine_subbasis_aliases.items():
            if mode in aliases:
                mode = key
                break
    if mode not in affine_subbasis_choices:
        raise ValueError('mode must be one of {}.'
                         .format(affine_subbasis_choices))

    # Compute the basis

    if mode == 'T':
        basis = np.zeros((dim, dim + 1, dim + 1), dtype=dtype)
        for i in range(dim):
            basis[i, i, dim] = 1

    elif mode == 'Z':
        basis = np.zeros((dim, dim + 1, dim + 1), dtype=dtype)
        for i in range(dim):
            basis[i, i, i] = 1

    elif mode == 'Z0':
        basis = np.zeros((dim - 1, dim + 1), dtype=dtype)
        for i in range(dim -1):
            basis[i, i] = 1
            basis[i, i + 1] = -1
        # Orthogonalise numerically.
        u, s, _ = np.linalg.svd(basis)
        s = s[..., None]
        basis = np.matmul(u.T, basis) / s
        diagbasis = basis
        basis = np.zeros(basis.shape + (dim+1), dtype=dtype)
        basis[..., np.arange(dim+1), np.arange(dim+1)] = diagbasis

        # TODO:
        #   Is there an analytical form?
        #   I would say yes. It seems that we have (I only list the diagonals):
        #       2D: [[a, -a]]           with a = 1/sqrt(2)
        #       3D: [[a, 0, -a],        with a = 1/sqrt(2)
        #            [b, -2*b, b]]           b = 1/sqrt(6)
        #       4D: [[a, -b, b, -a],
        #            [c, -c, -c, c],    with c = 1/sqrt(4)
        #            [b, a, -a, -b]]         a = ?, b = ?

    elif mode == 'I':
        basis = np.eye(dim + 1, dtype=dtype)[None, ...]
        basis /= pymath.sqrt(dim)
        basis[:, dim, dim] = 0

    elif mode == 'R':
        nb_rot = dim * (dim - 1) // 2
        basis = np.zeros((nb_rot, dim + 1, dim + 1), dtype=dtype)
        k = 0
        isqrt2 = 1 / pymath.sqrt(2)
        for i in range(dim):
            for j in range(i + 1, dim):
                basis[k, i, j] = isqrt2
                basis[k, j, i] = -isqrt2
                k += 1

    elif mode == 'S':
        nb_shr = dim * (dim - 1) // 2
        basis = np.zeros((nb_shr, dim + 1, dim + 1), dtype=dtype)
        k = 0
        isqrt2 = 1 / pymath.sqrt(2)
        for i in range(dim):
            for j in range(i + 1, dim):
                basis[k, i, j] = isqrt2
                basis[k, j, i] = isqrt2
                k += 1

    elif mode in ('sc', 'classic shear', 'classic shears'):
        nb_shr = dim * (dim - 1) // 2
        basis = np.zeros((nb_shr, dim + 1, dim + 1), dtype=dtype)
        k = 0
        for i in range(dim):
            for j in range(i + 1, dim):
                basis[k, i, j] = 1
                k += 1

    else:
        # We should never reach this (a test was performed earlier)
        raise ValueError

    # Select subcomponents of the basis
    if sub is not None:
        try:
            sub = list(sub)
        except TypeError:
            sub = [sub]
        basis = basis[sub]

    return basis


affine_basis_choices = ('T', 'SO', 'SE', 'D', 'CSO', 'SL', 'GL+', 'Aff+')
affine_basis_aliases = {
    'T': ['translation', 'translations'],
    'SO': ['rot', 'rotation', 'rotations'],
    'SE': ['rigid'],
    'D': ['dilation', 'dilations'],
    'CSO': ['similitude', 'similitudes'],
    'SL': [],
    'GL+': ['linear'],
    'Aff+': ['affine'],
}
affine_basis_components = {
    'T': ['T'],
    'SO': ['R'],
    'SE': ['T', 'R'],
    'D': ['R', 'I'],
    'CSO': ['T', 'R', 'I'],
    'SL': ['R', 'Z0', 'S'],
    'GL+': ['R', 'Z', 'S'],
    'Aff+': ['T', 'R', 'Z', 'S'],
}


def affine_basis(group='SE', dim=3, dtype=None):
    """Generate a basis set for the algebra of some (Lie) group of matrices.

    The basis is returned in homogeneous coordinates, even if
    the group does not require translations. To extract the linear
    part of the basis: lin = basis[:-1, :-1].

    This function focuses on 'classic' Lie groups. Note that, while it
    is commonly used in registration software, we do not have a
    "9-parameter affine" (translations + rotations + zooms),
    because such transforms do not form a group; that is, their inverse
    may contain shears.

    Parameters
    ----------
    group : {'T', 'SO', 'SE', 'D', 'CSO', 'SL', 'GL+', 'Aff+'}, default='SE'
        Group that should be encoded by the basis set:
            * 'T'    or 'translation' : Translations
            * 'SO'   or 'rotation'    : Special Orthogonal (rotations)
            * 'SE'   or 'rigid'       : Special Euclidean (translations + rotations)
            * 'D'    or 'dilation'    : Dilations (translations + isotropic scalings)
            * 'CSO'  or 'similitude'  : Conformal Special Orthogonal
                                        (translations + rotations + isotropic scalings)
            * 'SL'                    : Special Linear (rotations + isovolumic zooms + shears)
            * 'GL+'  or 'linear'      : General Linear [det>0] (rotations + zooms + shears)
            * 'Aff+' or 'affine'      : Affine [det>0] (translations + rotations + zooms + shears)
    dim : {1, 2, 3}, default=3
        Dimension
    dtype : torch.dtype, optional
        Data type of the returned array

    Returns
    -------
    basis : (F, dim+1, dim+1) tensor[dtype]
        Basis set, where ``F`` is the number of basis functions.

    """
    # TODO:
    # - other groups?

    # Authors
    # -------
    # .. John Ashburner <j.ashburner@ucl.ac.uk> : original Matlab code
    # .. Yael Balbastre <yael.balbastre@gmail.com> : Python code

    if group not in affine_basis_choices:
        for key, aliases in affine_basis_aliases.items():
            if group in aliases:
                group = key
                break
    if group not in affine_basis_choices:
        raise ValueError('group must be one of {}.'
                         .format(affine_basis_choices))

    subbases = [affine_subbasis(sub, dim, dtype=dtype)
                for sub in affine_basis_components[group]]
    return subbases[0] if len(subbases) == 1 else np.concatenate(subbases)


def affine_basis_size(group, dim=3):
    """Return the number of parameters in a given group."""

    if group not in affine_basis_choices:
        for key, aliases in affine_basis_aliases.items():
            if group in aliases:
                group = key
                break
    if group not in affine_basis_choices:
        raise ValueError('group must be one of {}.'
                         .format(affine_basis_choices))
    if group == 'T':
        return dim
    elif group == 'SO':
        return dim * (dim - 1) // 2
    elif group == 'SE':
        return dim + dim * (dim - 1) // 2
    elif group == 'D':
        return dim + 1
    elif group == 'CSO':
        return dim + dim * (dim - 1) // 2 + 1
    elif group == 'SL':
        return (dim + 1) * (dim - 1)
    elif group == 'GL+':
        return dim * dim
    elif group == 'Aff+':
        return (dim + 1) * dim


def affine_subbasis_size(basis, dim=3):
    """Return the number of parameters in a given group."""

    if basis not in affine_subbasis_choices:
        for key, aliases in affine_subbasis_aliases.items():
            if basis in aliases:
                basis = key
                break
    if basis not in affine_subbasis_choices:
        raise ValueError('basis must be one of {}.'
                         .format(affine_subbasis_choices))
    if basis in ('T', 'Z'):
        return dim
    elif basis in ('S', 'SC', 'R'):
        return dim * (dim - 1) // 2
    elif basis == 'Z0':
        return dim - 1
    elif basis == 'I':
        return 1


def build_affine_basis(*basis, dim=None, dtype=None):
    """Transform Affine Lie bases into tensors.

    Signatures
    ----------
    basis = build_affine_basis(basis)
    basis1, basis2, ... = build_affine_basis(basis1, basis2, ...)

    Parameters
    ----------
    *basis : basis_like or sequence[basis_like]
        A basis_like is a str or ([F], D+1, D+1) tensor_like that
        describes a basis. If several arguments are provided, each one
        is built independently.
    dim : int, optional
        Dimensionality. If None, infer.
    dtype : torch.dtype, optional
        Output data type

    Returns
    -------
    *basis : (F, D+1, D+1) tensor
        Basis sets.

    """
    if basis and (isinstance(basis[-1], int) or basis[-1] is None):
        *basis, dim = basis
    opts = dict(dtype=dtype, dim=dim)
    bases = [_build_affine_basis(b, **opts) for b in basis]
    return bases[0] if len(bases) == 1 else tuple(bases)


def _build_affine_basis(basis, dim=None, dtype=None):
    """Actual implementation: only one basis set."""

    # Helper to convert named bases to matrices
    def name_to_basis(name, dim, dtype):
        basename = name.split('[')[0]
        if basename in affine_subbasis_choices:
            return affine_subbasis(name, dim, dtype=dtype)
        elif basename in affine_basis_choices:
            return affine_basis(name, dim, dtype=dtype)
        else:
            raise ValueError('Unknown basis name {}.'.format(basename))

    # make list
    basis = ensure_list(basis)
    built_bases = [b for b in basis if not isinstance(b, str)]

    # check dimension
    dims = [dim] if dim else []
    dims = dims + [b.shape[-1] - 1 for b in built_bases]
    dims = set(dims)
    if not dims:
        dim = 3
    elif len(dims) > 1:
        raise ValueError('Dimension not consistent across bases.')
    else:
        dim = dims.pop()

    # build bases
    basis0 = basis
    basis = []
    for b in basis0:
        if isinstance(b, str):
            b = name_to_basis(b, dim, dtype)
        else:
            b = np.asanyarray(b)
        while b.ndim < 3:
            b = b[None]
        if b.shape[-2] != b.shape[-1] or b.ndim != 3:
            raise ValueError('Expected basis with shape (B, D+1, D+1) '
                             'but got {}'.format(tuple(b.shape)))
        basis.append(b)
    basis = np.concatenate(basis, axis=0)

    return basis



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math


def hash_function(state, value):
    """Python implementation of hashCombine() function
    in src/gpu_hash/hash_function.ts,
    a modified murmur hash
    """
    k1 = 0xCC9E2D51
    k2 = 0x1B873593
    state = state & 0xFFFFFFFF
    value = (value * k1) & 0xFFFFFFFF
    value = ((value << 15) | value >> 17) & 0xFFFFFFFF
    value = (value * k2) & 0xFFFFFFFF
    state = (state ^ value) & 0xFFFFFFFF
    state = ((state << 13) | state >> 19) & 0xFFFFFFFF
    state = ((state * 5) + 0xE6546B64) & 0xFFFFFFFF
    return state


def hsv_to_rgb(h, s, v):
    """Convert H,S,V values to RGB values.
    Python implementation of hsvToRgb in src/util/colorspace.ts"""
    h *= 6
    hue_index = math.floor(h)
    remainder = h - hue_index
    val1 = v * (1 - s)
    val2 = v * (1 - (s * remainder))
    val3 = v * (1 - (s * (1 - remainder)))
    hue_remainder = hue_index % 6
    if hue_remainder == 0:
        return (v, val3, val1)
    elif hue_remainder == 1:
        return (val2, v, val1)
    elif hue_remainder == 2:
        return (val1, v, val3)
    elif hue_remainder == 3:
        return (val1, val2, v)
    elif hue_remainder == 4:
        return (val3, val1, v)
    elif hue_remainder == 5:
        return (v, val1, val2)


def pack_color(rgb_vec):
    """Returns an integer formed
    by concatenating the channels of the input color vector.
    Python implementation of packColor in src/util/color.ts
    """
    result = 0
    for i in range(len(rgb_vec)):
        result = ((result << 8) & 0xFFFFFFFF) + min(
            255, max(0, round(rgb_vec[i] * 255))
        )
    return result


def hex_string_from_segment_id(color_seed, segment_id):
    """Return the hex color string for a segment
    given a color seed and the segment id"""
    segment_id = int(segment_id)  # necessary since segment_id is 64 bit originally
    result = hash_function(state=color_seed, value=segment_id)
    newvalue = segment_id >> 32
    result2 = hash_function(state=result, value=newvalue)
    c0 = (result2 & 0xFF) / 255.0
    c1 = ((result2 >> 8) & 0xFF) / 255.0
    h = c0
    s = 0.5 + 0.5 * c1
    v = 1.0
    rgb = hsv_to_rgb(h, s, v)
    packed_color = pack_color(rgb_vec=rgb)
    hex_string = format(packed_color, "x")
    """ Zero pad the hex string if less than 6 characeters """
    if len(hex_string) < 6:
        hex_string = "0" * (6 - len(hex_string)) + hex_string
    hex_string = "#" + hex_string
    return hex_string
